# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Classes that represent different aspects of a diagram."""
from __future__ import annotations

__all__ = [
    "Diagram",
    "DiagramElement",
    "Box",
    "Circle",
    "Edge",
    "RoutingStyle",
]

import collections.abc as cabc
import enum
import logging
import math
import os
import typing as t

from capellambse import aird, helpers

LOGGER = logging.getLogger(__name__)

SNAPPING = "AIRD_NOSNAP" not in os.environ

DiagramElement = t.Union["Box", "Edge", "Circle"]
_StyleOverrides = t.MutableMapping[
    str, t.Union[str, aird.RGB, t.MutableSequence[t.Union[str, aird.RGB]]]
]


class RoutingStyle(enum.Enum):
    OBLIQUE = enum.auto()
    MANHATTAN = enum.auto()
    TREE = enum.auto()


class Box:
    """A Box.  Some may call it rectangle."""

    # The number of pixels that a port hangs over its parent's border
    PORT_OVERHANG = 2
    CHILD_MARGIN = 2

    JSON_TYPE = "box"

    pos = aird.Vec2Property()
    minsize = aird.Vec2Property()
    maxsize = aird.Vec2Property()

    def __init__(
        self,
        pos: aird.Vec2ish,
        size: aird.Vec2ish,
        *,
        label: Box | str | None = None,
        uuid: str | None = None,
        parent: Box | None = None,
        collapsed: bool = False,
        minsize: aird.Vec2ish = aird.Vector2D(0, 0),
        maxsize: aird.Vec2ish = aird.Vector2D(math.inf, math.inf),
        context: cabc.Iterable[str] | None = None,
        port: bool = False,
        features: cabc.MutableSequence[str] | None = None,
        styleclass: str | None = None,
        styleoverrides: _StyleOverrides | None = None,
        hidelabel: bool = False,
        hidden: bool = False,
    ) -> None:
        """Create a new box.

        Parameters
        ----------
        pos
            A Vector2D describing the spatial position.
        size
            A Vector2D describing the box' size.  If one or both of its
            components are 0, it/they will be calculated based on the
            Box' label text and contained children.
        label
            This box' label text.
        uuid
            UUID of the semantic element this box represents.
        parent
            This box' parent box.
        collapsed
            Collapse this box and hide all its children.  Note that
            setting this flag does not change the box' size.
        minsize
            When dynamically calculating Box size, the minimum size it
            should have.  Default: zero.
        maxsize
            When dynamically calculating Box size, the maximum size it
            can have.  Default: inifite.
        context
            A list of UUIDs of objects in this box' context.  This
            includes children and associated edges.
        port
            Flag this box as a port.  Affects how context is added.
        features
            Certain classes of Box (like ``Class``) have features, which
            is a list of strings that will be displayed inside the Box,
            separated from the label by a horizontal line.
        styleclass
            The CSS style class to use.
        styleoverrides
            A dict of CSS properties to override.
        hidelabel
            Set to True to skip drawing this box' label.
        hidden
            Set to True to skip drawing this entire box.
        """
        self.uuid = uuid
        self.pos = pos
        self.size = aird.Vector2D(*size)
        self.minsize = minsize
        self.maxsize = maxsize
        self.label: Box | str | None = label
        self.collapsed: bool = collapsed
        self.features: cabc.MutableSequence[str] | None = features

        self.styleclass: str | None = styleclass
        self.styleoverrides: _StyleOverrides = styleoverrides or {}
        self.hidelabel: bool = hidelabel
        self.hidden = hidden

        self.children: cabc.MutableSequence[DiagramElement] = []
        self.context: set[str] = set(context) if context else set()
        self.port: bool = port

        self._parent: Box | None = None
        self.parent = parent

    def create_portlabel(
        self, labeltext: str, margin: float | int = 2
    ) -> None:
        """Add a label to a port box.

        The port that this is called for must be snapped to its parent's
        left or right side.

        Parameters
        ----------
        labeltext
            The label text.
        margin
            Space between the label text and the port Box.
        """
        if self.parent is None:
            raise ValueError("Ports must have a parent")
        if (
            self.pos.x > self.parent.pos.x
            and self.pos.x + self.size.x
            < self.parent.pos.x + self.parent.size.x
        ):
            raise ValueError(
                "Ports must be attached to their parents' side to get labels"
            )

        leftside = self.pos.x < self.parent.pos.x

        text_extent = aird.Vector2D(*helpers.get_text_extent(labeltext)) + (
            2 * margin,
            2 * margin,
        )
        text_pos = self.pos
        text_pos += (
            self.size.x + margin if leftside else -text_extent.x - margin,
            (self.size.y - text_extent.y) // 2,
        )
        self.label = type(self)(text_pos, text_extent, label=labeltext)

    def snap_to_parent(self) -> None:
        """Snap this Box into the constraints set by its parent.

        If this Box is a port, ensure it lines up with the parent's
        border, keeping an overhang of 2px.

        Otherwise, ensures that this Box will not overflow out of the
        parent's border, keeping a padding of 2px.
        """
        if not SNAPPING or self.parent is None:
            return

        if not isinstance(self.parent, Box):
            raise TypeError(
                f"Can only snap to parent Boxes, not {type(self.parent).__name__}"
            )

        if self.port:
            padding = (self.parent.PORT_OVERHANG, self.parent.PORT_OVERHANG)
            midbox_tl = self.parent.pos + self.size / 2 - padding
            midbox_br = (
                self.parent.pos + self.parent.size - self.size / 2 + padding
            )
            midbox = aird.Box(midbox_tl, midbox_br - midbox_tl)
            mid = self.pos + self.size / 2
            newmid = midbox.vector_snap(mid)
            self.pos += newmid - mid
        else:
            padding = (self.parent.CHILD_MARGIN, self.parent.CHILD_MARGIN)
            minpos = self.parent.pos + padding
            self.pos = aird.Vector2D(
                max(self.pos.x, minpos.x), max(self.pos.y, minpos.y)
            )
            maxsize = self.parent.pos + self.parent.size - self.pos - padding
            newsize = aird.Vector2D(
                min(self.size.x, maxsize.x), min(self.size.y, maxsize.y)
            )
            if newsize <= (0, 0):
                newsize = aird.Vector2D(0, 0)
                LOGGER.warning(
                    "Box %r (%s) has zero size after snapping",
                    self.label,
                    self.uuid,
                )
            self.size = aird.Vector2D(
                newsize.x if self._size.x > 0 else 0,
                newsize.y if self._size.y > 0 else 0,
            )

    def add_context(self, uuid: str) -> None:
        """Add a UUID as context for this box.

        The context will bubble to the immediate parent if this box is a
        port.
        """
        self.context.add(uuid)
        if self.port and self.parent is not None:
            self.parent.add_context(uuid)

    def vector_snap(
        self,
        point: aird.Vec2ish,
        *,
        source: aird.Vec2ish | None = None,
        style: RoutingStyle = RoutingStyle.OBLIQUE,
    ) -> aird.Vector2D:
        """Snap the ``point`` into this Box, coming from ``source``."""
        if not isinstance(point, aird.Vector2D):
            point = aird.Vector2D(*point)

        if not SNAPPING:
            return point

        if source is None:
            source = point
        elif not isinstance(source, aird.Vector2D):
            source = aird.Vector2D(*source)

        if style is RoutingStyle.OBLIQUE:
            return self.__vector_snap_oblique(source)
        if style is RoutingStyle.MANHATTAN:
            return self.__vector_snap_manhattan(point, point - source)
        if style is RoutingStyle.TREE:
            return self.__vector_snap_tree(point, point - source)
        raise ValueError(f"Unsupported routing style: {style}")

    def __vector_snap_oblique(self, source: aird.Vector2D) -> aird.Vector2D:
        angle = self.size.angleto(source - self.center)
        alpha = 2 * self.size.angleto((1, 0))
        assert alpha > 0
        alpha_prime = math.pi - alpha

        if 0 < angle < alpha:
            return aird.line_intersect(
                (self.center, source),
                (self.pos + self.size @ (1, 0), self.pos + self.size),
            )
        elif -alpha_prime < angle <= 0:
            return aird.line_intersect(
                (self.center, source),
                (self.pos + self.size @ (0, 1), self.pos + self.size),
            )
        elif alpha < angle < math.pi:
            return aird.line_intersect(
                (self.center, source),
                (self.pos, self.pos + self.size @ (1, 0)),
            )
        else:
            return aird.line_intersect(
                (self.center, source),
                (self.pos, self.pos + self.size @ (0, 1)),
            )

    def __vector_snap_manhattan(
        self, point: aird.Vector2D, direction: aird.Vector2D
    ) -> aird.Vector2D:
        direction = direction.closestaxis()
        if direction.x:
            if point.y < self.pos.y:
                return self.pos + self.size @ (0.5, 0)
            elif point.y > self.pos.y + self.size.y:
                return self.pos + self.size @ (0.5, 1)
            elif self.port:
                return self.pos + self.size @ (direction.x < 0, 0.5)
            else:
                return aird.Vector2D(
                    self.pos.x + self.size.x * (direction.x < 0),
                    point.y,
                )
        else:
            if point.x < self.pos.x:
                return self.pos + self.size @ (0, 0.5)
            elif point.x > self.pos.x + self.size.x:
                return self.pos + self.size @ (1, 0.5)
            elif self.port:
                return self.pos + self.size @ (0.5, direction.y < 0)
            else:
                return aird.Vector2D(
                    point.x,
                    self.pos.y + self.size.y * (direction.y < 0),
                )

    def __vector_snap_tree(
        self, point: aird.Vector2D, direction: aird.Vector2D
    ) -> aird.Vector2D:
        if direction == aird.Vector2D(0.0, 0.0):
            if self.center.x < point.x:
                return aird.Vector2D(point.x - 1, point.y)
            else:
                return aird.Vector2D(point.x + 1, point.y)

        if self.port:
            if (
                direction.y < 0.0
                or math.isclose(direction.y, 0.0)
                and point.y != self.pos.y
            ):
                return self.pos + self.size @ (0.5, 1.0)
            else:
                return self.pos + self.size @ (0.5, 0.0)
        else:
            if (
                direction.y < 0.0
                or math.isclose(direction.y, 0.0)
                and point.y != self.pos.y
            ):
                return aird.Vector2D(point.x, self.pos.y + self.size.y)
            else:
                return aird.Vector2D(point.x, self.pos.y)

    def move(self, offset: aird.Vector2D, *, children: bool = True) -> None:
        """Move the box by the specified offset.

        Parameters
        ----------
        children
            Recursively move children as well.  If False, the positions
            of children need to be adjusted separately.
        """
        self.pos += offset
        if isinstance(self.label, Box):
            self.label.move(offset, children=children)

        if children:
            for child in self.children:
                child.move(offset, children=True)

    @property
    def size(self) -> aird.Vector2D:
        """Return the size of this Box."""
        width, height = self._size
        needwidth = width <= 0
        needheight = height <= 0
        if not needwidth and not needheight:
            return self._size

        if isinstance(self.label, str):
            # pylint: disable=unpacking-non-sequence  # false-positive
            pad_w, pad_h = self.padding * 2  # Pad on all four sides

            # Fill in missing box size fields based on label text extent
            label_extent = helpers.get_text_extent(
                self.label + "\n" + "\n".join(self.features)
                if self.features
                else self.label,
                self.maxsize.x if needwidth else width - pad_w,
            )

            if needwidth:
                width = math.ceil(label_extent[0]) + pad_w
            if needheight:
                height = min(
                    self.maxsize.y, math.ceil(label_extent[1]) + pad_h
                )
        for child in self.children:
            if not child.hidden:
                cbound = child.bounds
                if needwidth:
                    width = max(
                        width,
                        cbound.pos.x
                        + cbound.size.x
                        + (-self.PORT_OVERHANG if child.port else 5)
                        - self.pos.x,
                    )
                    width = min(self.maxsize.x, width)
                if needheight:
                    height = max(
                        height,
                        cbound.pos.y
                        + cbound.size.y
                        + (-self.PORT_OVERHANG if child.port else 5)
                        - self.pos.y,
                    )
                    height = min(self.maxsize.y, height)

        # Features divider line
        if (
            self.features
            or self.styleclass == "Class"
            or self.styleclass == "Enumeration"
        ):
            height += 24

        width = max(self.minsize.x, width)
        height = max(self.minsize.y, height)
        return aird.Vector2D(width, height)

    @size.setter
    def size(self, new_size: aird.Vec2ish) -> None:
        if isinstance(new_size, aird.Vector2D):
            self._size = new_size
        else:
            self._size = aird.Vector2D(*new_size)

    @property
    def bounds(self) -> Box:
        """Calculate the bounding box of this Box."""
        minx, miny = self.pos
        maxx, maxy = self.pos + self.size
        if isinstance(self.label, Box) and not self.hidelabel:
            minx = min(minx, self.label.pos.x)
            miny = min(miny, self.label.pos.y)
            maxx = max(maxx, self.label.pos.x + self.label.size.x)
            maxy = max(maxy, self.label.pos.y + self.label.size.y)
        return Box((minx, miny), (maxx - minx, maxy - miny))

    @property
    def padding(self) -> aird.Vector2D:
        """Return the horizontal and vertical padding of label text."""
        if self.styleclass and self.styleclass.endswith("Annotation"):
            return aird.Vector2D(0, 0)
        return aird.Vector2D(10, 5)

    @property
    def hidden(self) -> bool:
        """Return whether to skip this Box during rendering."""
        hidden = self._hidden
        if self.parent is not None:
            hidden |= self.parent.hidden or self.parent.collapsed
        return hidden

    @hidden.setter
    def hidden(self, hide: bool) -> None:
        self._hidden = hide

    @property
    def center(self) -> aird.Vector2D:
        """The center point of this Box."""
        return self.pos + self.size / 2

    @property
    def parent(self) -> Box | None:
        """The parent element of this Box."""
        return self._parent

    @parent.setter
    def parent(self, parent: Box | None) -> None:
        self._parent, prev_parent = parent, self._parent
        if prev_parent is not None and self in prev_parent.children:
            prev_parent.children.remove(self)
        if self._parent is not None:
            self._parent.children.append(self)

        self.snap_to_parent()

    def __str__(self) -> str:
        if self.label is None:
            label = ""
        elif isinstance(self.label, str):
            label = self.label
        else:
            assert isinstance(self.label.label, str)
            label = self.label.label
        if label:
            label = " " + repr(label)
        return (
            f"{self.styleclass if self.styleclass else 'Box'}{label}"
            f" at {self.pos}, size {self.size}"
            + (f" with {len(self.features)} features" if self.features else "")
        )

    def __repr__(self) -> str:
        return "".join(
            [
                f"{type(self).__name__}({self.pos!r}, {self.size!r}",
                f", label={self.label!r}" if self.label else "",
                f", uuid={self.uuid!r}" if self.uuid is not None else "",
                (
                    f", parent=<UUID={self.parent.uuid}>"
                    if self.parent is not None
                    else ""
                ),
                (
                    f", context={{{', '.join(repr(c) for c in sorted(self.context))}}}"
                    if self.context
                    else ""
                ),
                f", features={self.features!r}" if self.features else "",
                f", styleclass={self.styleclass!r}" if self.styleclass else "",
                ", hidelabel=True" if self.hidelabel else "",
                ", hidden=True" if self.hidden else "",
                ")",
            ]
        )


class Edge(aird.Vec2List):
    """An edge in the diagram.

    An Edge consists of a series of points that are traversed in order.
    Each point is given as Vector2D containing absolute coordinates.  At
    least two points are required.
    """

    JSON_TYPE = "edge"

    collapsed = False
    context: cabc.Set[str] = frozenset()
    hidelabel = False
    port = False

    def __init__(
        self,
        points: cabc.Iterable[aird.Vec2ish],
        *,
        labels: cabc.MutableSequence[Box] | None = None,
        uuid: str | None = None,
        source: DiagramElement | None = None,
        target: DiagramElement | None = None,
        styleclass: str | None = None,
        styleoverrides: _StyleOverrides | None = None,
        hidden: bool = False,
    ):
        """Construct an Edge.

        Parameters
        ----------
        source
            The source diagram element of this Edge.
        target
            The target diagram element of this Edge.
        labels
            Labels for this Edge. Each label is a ``Box`` with ``pos``,
            ``size`` and a simple ``str`` label. Other configurations of
            Boxes are not supported. The ``hidden`` flag is honored
            during rendering calculations.
        points
            A list of ``Vector2D``s with the (absolute) points this edge
            follows.
        uuid
            UUID of the semantic element this Edge represents.
        styleclass
            The CSS style class to use.
        styleoverrides
            A dict of CSS properties to override.
        hidden
            True to skip drawing this edge entirely.
        """
        super().__init__(points)
        if len(self) < 2:
            raise ValueError(
                "At least two points are required"
                + (f" for edge with uuid {uuid}" if uuid else "")
            )

        self.source = source
        self.target = target

        self.uuid = uuid
        self.styleclass = styleclass
        self.styleoverrides = styleoverrides or {}

        self.labels = labels or []
        self.hidden = hidden

    def add_context(self, uuid: str) -> None:
        """Do nothing."""
        # Edges don't save context

    def vector_snap(
        self,
        vector: aird.Vec2ish,
        *,
        source: aird.Vec2ish,
        style: RoutingStyle = RoutingStyle.OBLIQUE,
    ) -> aird.Vector2D:
        """Snap the ``vector`` onto this Edge."""
        del source, style

        if not isinstance(vector, aird.Vector2D):
            vector = aird.Vector2D(*vector)

        if not SNAPPING:
            return vector

        epoint: aird.Vector2D  # Point on this Edge
        edist = math.inf  # Squared distance to that point

        for i in range(len(self.points) - 1):
            segment = self.points[i + 1] - self.points[i]
            seg_norm = segment.normalized
            dist = (vector - self.points[i]) * seg_norm

            # Clamp into the segment
            seglen = segment.length
            if dist > seglen:
                dist = seglen
            elif dist < 0:
                dist = 0

            new_epoint = self.points[i] + seg_norm * dist
            new_edist = (new_epoint - vector).sqlength

            if new_edist < edist:
                epoint = new_epoint
                edist = new_edist

        return epoint

    def move(self, offset: aird.Vector2D, *, children: bool = True) -> None:
        """Move all points of this edge by the specified offset."""
        del children
        for i, _ in enumerate(self):
            self[i] += offset

    @property
    def bounds(self) -> Box:
        """Calculate the bounding Box of this Edge."""
        labels = [i for i in self.labels if not i.hidden]
        if labels:
            minx = min(i.pos.x for i in labels)
            miny = min(i.pos.y for i in labels)
            maxx = max(i.pos.x + i.size.x for i in labels)
            maxy = max(i.pos.y + i.size.y for i in labels)
        else:
            minx = miny = math.inf
            maxx = maxy = -math.inf

        for point in self.points:
            if point.x < minx:
                minx = point.x
            if point.y < miny:
                miny = point.y
            if point.x > maxx:
                maxx = point.x
            if point.y > maxy:
                maxy = point.y

        topleft = aird.Vector2D(minx, miny)
        bottomright = aird.Vector2D(maxx, maxy)
        return Box(topleft, bottomright - topleft)

    @property
    def hidden(self) -> bool:
        """Return whether to skip this Edge during rendering."""
        return (
            self._hidden
            or (self.source is not None and self.source.hidden)
            or (self.target is not None and self.target.hidden)
        )

    @hidden.setter
    def hidden(self, hide: bool) -> None:
        self._hidden = hide

    @property
    def points(self) -> aird.Vec2List:
        """Return an iterable over this edge's points."""
        return self

    @points.setter
    def points(self, newpoints: cabc.Iterable[aird.Vec2ish]) -> None:
        self[:] = newpoints

    @property
    def length(self) -> float:
        """Return length of this edge."""
        return sum((self[i - 1] - self[i]).length for i in range(1, len(self)))

    @property
    def center(self) -> aird.Vector2D:
        """Calculate the point in the middle of this edge."""
        half_length = self.length / 2
        distance = 0.0
        point = self[0]
        for i in range(1, len(self)):
            segment = self[i] - self[i - 1]
            if distance + segment.length > half_length:
                point += segment.normalized * (half_length - distance)
                break
            distance += segment.length
            point = self[i]
        return point

    def __str__(self) -> str:
        numpoints = len(self.points)
        if self.labels:
            label = self.labels[0].label or ""
        else:
            label = ""
        return (
            f"{self.styleclass or 'Edge'}"
            f"{f' {label!r}' if label else ''} between"
            f" {self[0]} and {self[-1]}"
            + (
                # pylint: disable=not-an-iterable  # false-positive
                f" via {', '.join(str(p) for p in self[1:-1])}"
                if numpoints > 2
                else ""
            )
        )

    def __repr__(self) -> str:
        return "".join(
            [
                f"{self.__class__.__name__}({list(self)!r}",
                f", labels={self.labels!r}" if self.labels else "",
                f", uuid={self.uuid!r}" if self.uuid else "",
                f", styleclass={self.styleclass!r}" if self.styleclass else "",
                ", hidden=True" if self.hidden else "",
                ")",
            ]
        )


class Circle:
    """Represents a circle."""

    JSON_TYPE = "circle"

    collapsed = False
    context: cabc.Set[str] = frozenset()
    hidelabel = True
    label = None
    port = False

    center = aird.Vec2Property()

    def __init__(
        self,
        center: aird.Vec2ish,
        radius: float | int,
        *,
        uuid: str | None = None,
        styleclass: str | None = None,
        styleoverrides: _StyleOverrides | None = None,
        hidden: bool = False,
    ):
        """Construct a Circle.

        Parameters
        ----------
        center
            The Circle's center point as Vector2D or 2-tuple.
        radius
            The Circle radius in pixels.
        uuid
            The Circle's unique identifier.
        styleclass
            The Circle's CSS class.
        styleoverrides
            Dict of CSS key/value pairs that override the class
            defaults.
        hidden
            True to skip drawing this Circle.
        """
        self.uuid = uuid
        self.center = center
        self.radius = radius

        self.styleclass = styleclass
        self.styleoverrides = styleoverrides or {}
        self.hidden = hidden

    def add_context(self, uuid: str) -> None:
        """Do nothing."""
        # Circles don't save context

    @property
    def bounds(self) -> Box:
        """Calculate the bounding box of this Circle."""
        return Box(
            self.center - (self.radius, self.radius),
            (self.radius * 2, self.radius * 2),
        )

    def move(self, offset: aird.Vec2ish, *, children: bool = True) -> None:
        """Move this Circle on the 2-dimensional plane."""
        del children
        self.center += offset

    def vector_snap(
        self,
        vector: aird.Vec2ish,
        *,
        source: aird.Vec2ish,
        style: RoutingStyle = RoutingStyle.OBLIQUE,
    ) -> aird.Vector2D:
        """Snap the ``vector`` onto this Circle, preferably in `direction`."""
        del style
        # TODO implement different routing styles than OBLIQUE

        if not isinstance(vector, aird.Vector2D):
            vector = aird.Vector2D(*vector)
        if not isinstance(source, aird.Vector2D):
            source = aird.Vector2D(*source)
        direction = source - vector
        if vector == self.center:
            vector = direction
        assert vector != (0, 0)

        return self.center + vector.normalized * self.radius

    def __str__(self) -> str:
        return (
            f"{f'{self.styleclass}-' if self.styleclass else ''}Circle"
            f"at {self.center} with radius {self.radius}px"
        )

    def __repr__(self) -> str:
        return "".join(
            (
                f"{self.__class__.__name__}({self.center!r}, {self.radius!r}",
                f", uuid={self.uuid!r}" if self.uuid else "",
                f", styleclass={self.styleclass!r}" if self.styleclass else "",
                (
                    f", styleoverrides={self.styleoverrides!r}"
                    if self.styleoverrides
                    else ""
                ),
                ", hidden=True" if self.hidden else "",
                ")",
            )
        )


class Diagram:
    """A complete diagram, including all elements required by it."""

    def __init__(
        self,
        name: str = "Untitled Diagram",
        viewport: Box | None = None,
        elements: cabc.Sequence[DiagramElement] | None = None,
        *,
        uuid: str | None = None,
        styleclass: str | None = None,
    ):
        """Construct a new diagram.

        Parameters
        ----------
        name
            The diagram's name.
        viewport
            A Box describing this diagram's viewport.
        elements
            A :class:`list` containing the diagram's initial elements.
        uuid
            The unique ID of this diagram.
        styleclass
            The diagram class.
        """
        self.name = name
        self.uuid = uuid
        self.styleclass = styleclass

        self.viewport = None
        self.__elements: list[DiagramElement] = []

        if elements is not None:
            for element in elements:
                self.add_element(element)
        if viewport is not None:
            self.viewport = Box(
                viewport.pos, viewport.size, styleclass="Viewport"
            )

    def add_element(
        self,
        element: DiagramElement,
        extend_viewport: bool = True,
        *,
        force: bool = False,
    ) -> None:
        """Add an element to this diagram.

        Parameters
        ----------
        element
            The element to add.
        extend_viewport
            True to automatically extend the diagram viewport so that
            the added element is fully visible.
        """
        if element.uuid is not None:
            if element.uuid in self:
                if force:
                    old_hidden = self[element.uuid].hidden
                    new_hidden = element.hidden
                    if new_hidden:
                        LOGGER.warning("Skipping hidden duplicate %s", element)
                        return
                    if not old_hidden:
                        LOGGER.warning("Skipping duplicate %s", element)
                        return
                    LOGGER.warning(
                        "Overwriting hidden duplicate %r with %r",
                        str(self[element.uuid]),
                        str(element),
                    )
                    self.__elements.remove(self[element.uuid])
                else:
                    raise ValueError(
                        f"Duplicate element UUID {element.uuid!r}"
                    )

        if extend_viewport:
            self.__extend_viewport(element.bounds)
        self.__elements.append(element)

    def calculate_viewport(self) -> None:
        """Recalculate the viewport so that all elements are contained."""
        minx = miny = math.inf
        maxx = maxy = -math.inf

        for elm in self:
            if elm.hidden:
                continue

            bounds = elm.bounds
            minx = min(minx, bounds.pos.x)
            miny = min(miny, bounds.pos.y)
            maxx = max(maxx, bounds.pos.x + bounds.size.x)
            maxy = max(maxy, bounds.pos.y + bounds.size.y)

        top_left = aird.Vector2D(minx, miny)
        bottom_right = aird.Vector2D(maxx, maxy)
        self.viewport = Box(
            top_left, bottom_right - top_left, styleclass="Viewport"
        )

    def normalize_viewport(
        self,
        offset: float | int | aird.Vec2ish = 0,
    ) -> None:
        """Normalize the viewport.

        This function moves all elements contained within this diagram
        so that the top left corner of the viewport is at (0, 0) (or the
        specified offset, if given).

        If a single value is given as offset, it is applied to both X
        and Y coordinates.
        """
        if isinstance(offset, aird.Vector2D):
            offsetvec = offset
        elif isinstance(offset, (list, tuple)):
            offsetvec = aird.Vector2D(*offset)
        else:
            offsetvec = aird.Vector2D(offset, offset)
        del offset

        if self.viewport is None:
            self.calculate_viewport()
        assert self.viewport is not None

        offsetvec -= self.viewport.pos

        for element in [self.viewport, *self.__elements]:
            element.move(offsetvec, children=False)

    def __extend_viewport(self, element: DiagramElement) -> None:
        """Extend the viewport so the given element fits in.

        If the element's bounding box already lies completely within the
        current viewport, this function does nothing.
        """
        bounds = element.bounds

        if self.viewport is None:
            self.viewport = Box(bounds.pos, bounds.size, styleclass="Viewport")
        else:
            top_left = aird.Vector2D(
                min(self.viewport.pos.x, bounds.pos.x),
                min(self.viewport.pos[1], bounds.pos[1]),
            )
            bottom_right = aird.Vector2D(
                max(
                    self.viewport.pos.x + self.viewport.size.x,
                    bounds.pos.x + bounds.size.x,
                ),
                max(
                    self.viewport.pos.y + self.viewport.size.y,
                    bounds.pos.y + bounds.size.y,
                ),
            )
            pos = top_left
            size = bottom_right - top_left
            self.viewport = Box(pos, size, styleclass="Viewport")

    def __len__(self) -> int:
        return len(self.__elements)

    def __getitem__(self, key: int | str) -> DiagramElement:
        if isinstance(key, int):  # sequence protocol
            return self.__elements[key]

        if isinstance(key, str):  # lookup by uuid
            for elm in self.__elements:
                if elm.uuid == key:
                    return elm
            raise KeyError(f"No element with uuid {key!r} in this diagram")

        raise TypeError(f"Cannot look up elements by {type(key).__name__!s}")

    def __iter__(self) -> cabc.Iterator[DiagramElement]:
        return iter(self.__elements)

    def __contains__(self, obj: str | DiagramElement) -> bool:
        if isinstance(obj, str):
            try:
                self[obj]
            except KeyError:
                return False
            else:
                return True
        else:
            return obj in self.__elements

    def __str__(self) -> str:
        return "".join(
            [
                self.styleclass or self.__class__.__name__,
                f" {self.name!r}",
                f" ({self.uuid!r})" if self.uuid else "",
                f" [{self.viewport!s}]\n",
                "\n".join(f"\t- {elm}" for elm in self)
                if self
                else "\t(empty)",
            ]
        )

    def __repr__(self) -> str:
        return "".join(
            [
                f"{self.__class__.__name__}("
                f"{self.name!r}, {self.viewport!r}, {self.__elements!r}",
                f", uuid={self.uuid!r}" if self.uuid else "",
                f", styleclass={self.styleclass!r}" if self.styleclass else "",
                ")",
            ]
        )

    def __iadd__(self, element: DiagramElement) -> Diagram:
        self.add_element(element)
        return self



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import collections.abc as cabc
import dataclasses
import json
import os
import pathlib
import typing as t

from capellambse.svg.drawing import LabelDict

ContentsDict = t.TypedDict(
    "ContentsDict",
    {
        "type": str,
        "id": str,
        "class": str,
        "points": t.List[t.List[int]],
        "x": float,
        "y": float,
        "width": float,
        "height": float,
        "label": t.Union[LabelDict, str],
    },
    total=False,
)


class SVGDiagram:
    """An SVG diagram that can be drawn on and serialized.

    SVG diagram object that takes the ``metadata`` of a diagram via the
    :class:`DiagramMetadata` and a list of objects that are dictionaries
    describing the components to be drawn on the diagram canvas of type
    :class:`Drawing`.

    Example of expected json-file/string::

        {
            "name": "FA00 - Functional Architecture Example",
            "class": "LogicalArchitectureBlank",
            "x": 10,
            "y": 20,
            "width": 100,
            "height": 200,
            "contents": [
                {
                    "type": "box",
                    "id": "_ZNVPYDHuEeqOg4absf8kjA",
                    "class": "LogicalFunction",
                    "x": 150,
                    "y": 130,
                    "width": 101,
                    "height": 101,
                    "label": "example label"
                }
            ]
        }
    """

    def __init__(
        self,
        metadata: DiagramMetadata,
        objects: cabc.Sequence[ContentsDict],
    ) -> None:
        self.drawing = Drawing(metadata)
        for obj in objects:
            self.draw_object(obj)

    @classmethod
    def from_json(cls, jsonstring: str) -> SVGDiagram:
        """Create an SVGDiagram from the given JSON string.

        Parameters
        ----------
        jsonstring
            Json/dictionary in ``str`` format

        Returns
        -------
        diagram
            SVG diagram object
        """
        jsondict = json.loads(jsonstring)
        metadata = DiagramMetadata.from_dict(jsondict)
        return cls(metadata, jsondict["contents"])

    @classmethod
    def from_json_path(cls, path: str | os.PathLike) -> SVGDiagram:
        """Create an SVGDiagram from the given JSON file.

        Parameters
        ----------
        path
            path to .json file

        Returns
        -------
        diagram
            SVG diagram object
        """
        conf = pathlib.Path(path).read_text(encoding="utf-8")
        return cls.from_json(conf)

    def draw_object(self, obj: ContentsDict) -> None:
        self.drawing.draw_object(obj)

    def save_drawing(
        self, pretty: bool = False, indent: int = 2, filename: str = ""
    ) -> None:
        self.drawing.saveas(
            filename=filename or self.drawing.filename,
            pretty=pretty,
            indent=indent,
        )

    def to_string(self) -> str:
        return self.drawing.tostring()


DiagramMetadataDict = t.TypedDict(
    "DiagramMetadataDict",
    {
        "name": str,
        "contents": t.Sequence[ContentsDict],
        "x": float,
        "y": float,
        "width": float,
        "height": float,
        "class": t.Union[str, None],
    },
)


@dataclasses.dataclass
class DiagramMetadata:
    """Holds metadata about a diagram.

    The metadata of a diagram includes the diagram-name, ``(x, y)``
    position, ``(w, h)`` size, the viewbox string and the diagram class,
    e.g. ``LogicalArchitectureBlank``.
    """

    def __init__(
        self,
        pos: tuple[float, float],
        size: tuple[float, float],
        name: str,
        class_: str | None,
        **_kw: t.Any,
    ) -> None:
        # Add padding to viewbox to account for drawn borders
        self.pos: tuple[float, float] = tuple(i - 10 for i in pos)  # type: ignore[assignment]
        self.size: tuple[float, float] = tuple(i + 20 for i in size)  # type: ignore[assignment]
        self.viewbox = " ".join(map(str, self.pos + self.size))
        self.class_ = class_
        self.name = name

    @classmethod
    def from_dict(cls, data: DiagramMetadataDict) -> DiagramMetadata:
        name = data.get("name")
        if not name:
            raise TypeError("No diagram name defined")
        return cls(
            (data["x"], data["y"]),
            (data["width"], data["height"]),
            name,
            data["class"],
        )


from .drawing import Drawing



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Classes that allow access to diagrams in the model."""
from __future__ import annotations

import abc
import base64
import collections.abc as cabc
import importlib.metadata as imm
import logging
import operator
import traceback
import typing as t
import uuid

import markupsafe

import capellambse
from capellambse.loader import xmltools

from .. import aird, helpers, svg
from . import common as c
from . import modeltypes


@t.runtime_checkable
class DiagramFormat(t.Protocol):
    filename_extension: str

    @classmethod
    def convert(cls, diagram: aird.Diagram) -> t.Any:
        ...

    @classmethod
    def from_cache(cls, cache: bytes) -> t.Any:
        ...


DiagramConverter = t.Union[
    t.Callable[[aird.Diagram], t.Any],
    DiagramFormat,
]

LOGGER = logging.getLogger(__name__)


class UnknownOutputFormat(ValueError):
    """An unknown output format was requested for the diagram."""


class AbstractDiagram(metaclass=abc.ABCMeta):
    """Abstract superclass of model diagrams."""

    uuid: str
    """Unique ID of this diagram."""
    name: str
    """Human-readable name for this diagram."""
    target: c.GenericElement
    """This diagram's "target".

    The target of a diagram is usually:

    *   The model element which is the direct parent of all visible
        nodes **OR**
    *   The only top-level element in the diagram **OR**
    *   The element which is considered to be the "element of interest".
    """
    filters: cabc.MutableSet[str]
    """The filters that are activated for this diagram."""

    _model: capellambse.MelodyModel
    _render: aird.Diagram
    _error: BaseException
    _last_render_params: dict[str, t.Any] = {}
    """
    Additional rendering parameters for the cached rendered diagram.

    Rendering options for :class:`aird.Diagram`s. Handing over
    parameters that differ to these will force a fresh rendering of the
    diagram, flushing the cached diagram.

    The following parameters are currently supported:

        - ``sorted_exchangedItems`` (*bool*): Enable ExchangeItem
          sorting when rendering diagrams with active ExchangeItems
          filter (``show.exchange.items.filter``).
    """

    def __init__(self, model: capellambse.MelodyModel) -> None:
        self._model = model
        self._last_render_params = {}

    def __dir__(self) -> list[str]:
        return dir(type(self)) + [
            f"as_{i.name}"
            for i in imm.entry_points()["capellambse.diagram.formats"]
        ]

    def __getattr__(self, attr: str) -> t.Any:
        if attr.startswith("as_"):
            fmt = attr[len("as_") :]
            try:
                return self.render(fmt)
            except UnknownOutputFormat:
                raise
            except Exception as err:  # pylint: disable=broad-except
                if hasattr(self, "_error") and err is self._error:
                    err_img = self._render
                else:
                    err_img = self.__create_error_image("render", err)
                assert err_img is not None
                converter = _find_format_converter(fmt)
                if isinstance(converter, DiagramFormat):
                    return converter.convert(err_img)
                else:
                    return converter(err_img)
        return getattr(super(), attr)

    def __repr__(self) -> str:
        return f"<Diagram {self.name!r}>"

    def __html__(self) -> markupsafe.Markup:
        return (
            markupsafe.Markup("<figure>")
            + markupsafe.Markup(self.render("svg"))
            + markupsafe.Markup("<figcaption>")
            + self.name
            + markupsafe.Markup("</figcaption></figure>")
        )

    def _short_html_(self) -> markupsafe.Markup:
        return markupsafe.Markup(
            f"<b>{markupsafe.Markup.escape(self.name)}</b> "
            f"(uuid: {markupsafe.Markup.escape(self.uuid)})"
        )

    def _repr_mimebundle_(
        self,
        include: cabc.Container[str] | None = None,
        exclude: cabc.Container[str] | None = None,
        **_,
    ) -> tuple[dict[str, t.Any], dict[t.Any, t.Any]] | dict[str, t.Any] | None:
        if include is None:
            include = helpers.EverythingContainer()
        if exclude is None:
            exclude = ()

        formats: dict[str, DiagramConverter] = {}
        for conv in _iter_format_converters():
            mime = getattr(conv, "mimetype", None)
            if not mime or mime not in include or mime in exclude:
                continue

            # XXX Hack to fix diagram previews on Github-rendered notebooks
            if self._model.jupyter_untrusted and mime != "image/png":
                continue

            formats[mime] = conv
        if not formats:
            return None

        bundle: dict[str, t.Any] = {}
        for mime, conv in formats.items():
            try:
                bundle[mime] = self.__load_cache(conv)
            except KeyError:
                pass

        if bundle:
            return bundle

        render = self.__render_fresh({})
        for mime, conv in formats.items():
            try:
                if isinstance(conv, DiagramFormat):
                    bundle[mime] = conv.convert(render)
                else:
                    bundle[mime] = conv(render)
            except Exception:
                LOGGER.exception("Failed converting diagram with %r", conv)
        if not bundle:
            LOGGER.error("Failed converting diagram for MIME bundle")
            bundle["text/plain"] = repr(self)
        return bundle

    @property
    def nodes(self) -> c.MixedElementList:
        """Return a list of all nodes visible in this diagram."""
        allids = {e.uuid for e in self.render(None)}
        assert None not in allids
        elems = []
        for elemid in allids:
            assert elemid is not None
            try:
                elem = self._model._loader[elemid]
                elem = next(elem.iterchildren("target"))
                elem = self._model._loader.follow_link(
                    elem, elem.attrib["href"]
                )
            except (KeyError, StopIteration):  # pragma: no cover
                continue
            else:
                # Filter out visual-only elements that live in the
                # .aird / .airdfragment files
                frag = self._model._loader.find_fragment(elem)
                if frag.suffix not in {".aird", ".airdfragment"}:
                    elems.append(elem)

        return c.MixedElementList(self._model, elems, c.GenericElement)

    @t.overload
    def render(self, fmt: None, **params) -> aird.Diagram:
        ...

    @t.overload
    def render(self, fmt: str, **params) -> t.Any:
        ...

    def render(self, fmt: str | None, **params) -> t.Any:
        """Render the diagram in the given format."""
        if fmt is not None:
            conv = _find_format_converter(fmt)
        else:
            conv = lambda i: i

        try:
            return self.__load_cache(conv)
        except KeyError:
            pass

        render = self.__render_fresh(params)
        if isinstance(conv, DiagramFormat):
            return conv.convert(render)
        else:
            return conv(render)

    @abc.abstractmethod
    def _create_diagram(self, params: dict[str, t.Any]) -> aird.Diagram:
        """Perform the actual rendering of the diagram.

        This method should only be called by the public :meth:`render`
        method, as it handles caching of the results.
        """

    def __create_error_image(
        self, stage: str, error: Exception
    ) -> aird.Diagram:
        err_name = (
            "An error occured while rendering diagram\n"
            f"{self.name!r}\n"
            f"(in stage {stage!r})"
        )
        err_msg = (
            "Please report this error to your tools and methods team,"
            " and attach the following information:"
        )
        err_trace = "".join(
            traceback.format_exception(None, error, error.__traceback__)
        )

        diag = aird.Diagram("An error occured! :(")
        err_box = aird.Box(
            (200, 0),
            (350, 0),
            label=err_name,
            uuid="error",
            styleclass="Note",
            styleoverrides={
                "fill": aird.RGB(255, 0, 0),
                "text_fill": aird.RGB(255, 255, 255),
            },
        )
        diag.add_element(err_box, extend_viewport=False)
        info_box = aird.Box(
            (200, err_box.pos.y + err_box.size.y + 10),
            (350, 0),
            label=err_msg,
            uuid="info",
            styleclass="Note",
        )
        diag.add_element(info_box, extend_viewport=False)
        trace_box = aird.Box(
            (0, info_box.pos.y + info_box.size.y + 10),
            (750, 0),
            label=err_trace,
            uuid="trace",
            styleclass="Note",
        )
        diag.add_element(trace_box, extend_viewport=False)
        diag.calculate_viewport()
        return diag

    def __load_cache(self, converter: DiagramConverter):
        cache_handler = getattr(self._model, "_diagram_cache", None)
        cachedir = getattr(self._model, "_diagram_cache_subdir", None)
        if cache_handler is None or cachedir is None:
            raise KeyError(self.uuid)

        if not isinstance(converter, DiagramFormat):
            raise KeyError(self.uuid)

        try:
            ext = converter.filename_extension
            with cache_handler.open(cachedir / (self.uuid + ext)) as f:
                cache = f.read()
        except FileNotFoundError:
            raise KeyError(self.uuid) from None

        return converter.from_cache(cache)

    def __render_fresh(self, params: dict[str, t.Any]) -> aird.Diagram:
        # pylint: disable=broad-except
        if not hasattr(self, "_render") or self._last_render_params != params:
            self.invalidate_cache()
            try:
                self._render = self._create_diagram(params)
            except Exception as err:
                self._error = err
                self._render = self.__create_error_image("parse", self._error)

        if hasattr(self, "_error"):
            raise self._error
        return self._render

    def invalidate_cache(self) -> None:
        """Reset internal diagram cache."""
        try:
            del self._render
        except AttributeError:
            pass

        try:
            del self._error
        except AttributeError:
            pass


class Diagram(AbstractDiagram):
    """Provides access to a single diagram."""

    uuid: str = property(operator.attrgetter("_element.uid"))  # type: ignore[assignment]
    xtype = "viewpoint:DRepresentationDescriptor"
    name: str = property(operator.attrgetter("_element.name"))  # type: ignore[assignment]
    viewpoint: str = property(operator.attrgetter("_element.viewpoint"))  # type: ignore[assignment]
    target_uuid: str = property(lambda self: self.target.uuid)  # type: ignore[assignment]
    """Obsolete."""
    description = xmltools.HTMLAttributeProperty(
        "_element", "documentation", optional=True
    )

    _element: aird.DiagramDescriptor

    @classmethod
    def from_model(
        cls,
        model: capellambse.MelodyModel,
        descriptor: aird.DiagramDescriptor,
    ) -> Diagram:
        """Wrap a diagram already defined in the Capella AIRD."""
        self = cls.__new__(cls)
        self._model = model
        self._element = descriptor
        return self

    def __init__(self, **kw: t.Any) -> None:
        # pylint: disable=super-init-not-called
        raise TypeError("Cannot create a Diagram this way")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._model is other._model and self._element == other._element

    @property
    def target(self) -> c.GenericElement:  # type: ignore[override]
        return c.GenericElement.from_model(self._model, self._element.target)

    @property
    def type(self) -> modeltypes.DiagramType:
        """Return the type of this diagram."""
        sc = self._element.styleclass
        try:
            return modeltypes.DiagramType(sc)
        except ValueError:  # pragma: no cover
            LOGGER.warning("Unknown diagram type %r", sc)
            return modeltypes.DiagramType.UNKNOWN

    @property  # type: ignore[override]
    def filters(self) -> cabc.MutableSet[str]:  # type: ignore[override]
        """Return a set of currently activated filters on this diagram."""
        return aird.ActiveFilters(self._model, self)

    @filters.setter
    def filters(self, filters: cabc.Iterable[str]) -> None:
        self.filters.clear()
        for filter in filters:
            self.filters.add(filter)

    def _create_diagram(self, params: dict[str, t.Any]) -> aird.Diagram:
        return aird.parse_diagram(self._model._loader, self._element, **params)


class DiagramAccessor(c.Accessor):
    """Provides access to a list of diagrams below the specified viewpoint."""

    def __init__(
        self,
        viewpoint: str | None = None,
        *,
        cacheattr: str | None = None,
    ) -> None:
        super().__init__()
        self.cacheattr = cacheattr
        self.viewpoint = viewpoint

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        if self.viewpoint is None:
            descriptors = list(aird.enumerate_diagrams(obj._model._loader))
        else:
            descriptors = [
                d
                for d in aird.enumerate_diagrams(obj._model._loader)
                if d.viewpoint == self.viewpoint
            ]
        return c.CachedElementList(
            obj._model,
            descriptors,
            Diagram,
            cacheattr=self.cacheattr,
        )


class SVGFormat:
    """Convert the diagram to SVG."""

    filename_extension = ".svg"
    mimetype = "image/svg+xml"

    @staticmethod
    def convert(diagram: aird.Diagram) -> str:
        return convert_svgdiagram(diagram).to_string()

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


class PNGFormat:
    """Convert the diagram to PNG."""

    filename_extension = ".png"
    mimetype = "image/png"

    @staticmethod
    def convert(diagram: aird.Diagram) -> bytes:
        try:
            import cairosvg
        except OSError as error:
            raise RuntimeError(
                "Cannot import cairosvg. You are likely missing .dll's."
                " Please see the README for instructions."
            ) from error

        return cairosvg.svg2png(SVGFormat.convert(diagram))

    @staticmethod
    def from_cache(cache: bytes) -> bytes:
        return cache


def convert_svgdiagram(
    diagram: aird.Diagram,
) -> svg.generate.SVGDiagram:
    """Convert the diagram to a SVGDiagram."""
    jsondata = aird.DiagramJSONEncoder().encode(diagram)
    return svg.generate.SVGDiagram.from_json(jsondata)


class ConfluenceSVGFormat:
    """Convert the diagram to Confluence-style SVG."""

    filename_extension = ".svg"
    prefix = (
        f'<ac:structured-macro ac:macro-id="{uuid.uuid4()!s}" ac:name="html"'
        ' ac:schema-version="1">'
        f"<ac:plain-text-body><![CDATA["
    )
    postfix = "]]></ac:plain-text-body></ac:structured-macro>"

    @classmethod
    def convert(cls, diagram: aird.Diagram) -> str:
        return "".join((cls.prefix, SVGFormat.convert(diagram), cls.postfix))

    @classmethod
    def from_cache(cls, cache: bytes) -> str:
        return "".join((cls.prefix, cache.decode("utf-8"), cls.postfix))


class SVGDataURIFormat:
    filename_extension = ".svg"
    preamble = "data:image/svg+xml;base64,"

    @classmethod
    def convert(cls, diagram: aird.Diagram) -> str:
        payload = SVGFormat.convert(diagram)
        b64 = base64.standard_b64encode(payload.encode("utf-8"))  # type: ignore[attr-defined]
        return "".join((cls.preamble, b64.decode("ascii")))

    @classmethod
    def from_cache(cls, cache: bytes) -> str:
        b64 = base64.standard_b64encode(cache)
        return "".join((cls.preamble, b64.decode("ascii")))


class SVGInHTMLIMGFormat:
    filename_extension = ".svg"
    mimetype = "text/html"

    @staticmethod
    def convert(diagram: aird.Diagram) -> markupsafe.Markup:
        payload = SVGDataURIFormat.convert(diagram)
        return markupsafe.Markup(f'<img src="{payload}"/>')

    @staticmethod
    def from_cache(cache: bytes) -> str:
        payload = SVGDataURIFormat.from_cache(cache)
        return markupsafe.Markup(f'<img src="{payload}"/>')


class JSONFormat:
    filename_extension = ".json"

    @staticmethod
    def convert(diagram: aird.Diagram) -> str:
        return aird.DiagramJSONEncoder().encode(diagram)

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


class PrettyJSONFormat:
    filename_extension = ".json"
    mimetype = "application/json"

    @staticmethod
    def convert(diagram: aird.Diagram) -> str:
        return aird.DiagramJSONEncoder(indent=4).encode(diagram)

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


def _find_format_converter(fmt: str) -> DiagramConverter:
    try:
        return next(
            i.load()
            for i in imm.entry_points()["capellambse.diagram.formats"]
            if i.name == fmt
        )
    except StopIteration:
        raise UnknownOutputFormat(
            f"Unknown image output format {fmt}"
        ) from None


def _iter_format_converters() -> t.Iterator[DiagramConverter]:
    for ep in imm.entry_points()["capellambse.diagram.formats"]:
        try:
            conv = ep.load()
        except ImportError:
            pass
        else:
            yield conv



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import collections.abc as cabc
import dataclasses
import re
import typing as t

PLUGIN_PATTERN = re.compile(r"(.*?)(\d\.\d\.\d)?$")


@dataclasses.dataclass
class Version:
    """Capella xml-element/plugin version info about name and version."""

    plugin: str
    version: str | None = None

    @property
    def values(self) -> cabc.Iterator[str | None]:
        yield self.plugin
        yield self.version

    def __le__(self, other: float | int | str | Version) -> t.Any:
        if self.version is None:
            return False

        if isinstance(other, str):
            other = _tofloat(other)
        return other <= _tofloat(self.version)

    def __str__(self) -> str:
        return self.plugin + (self.version or "")


def _tofloat(other: str) -> float:
    """Change 1.x.y...str into 1.x float."""
    version = other.split(".")
    if len(version) > 1:
        version = version[:2]

    return float(".".join(version))


def yield_key_and_version_from_namespaces_by_plugin(
    plugin: str,
) -> cabc.Iterator[tuple[str, Version]]:
    """Yield namespace key and :class:`Version` tuple for plugin str.

    Parameters
    ----------
    plugin
        ``xsi:type`` string of xml-element/plugin

    Yields
    ------
    tuple
        If plugin in namespace values yield key value and Version
    """
    match = PLUGIN_PATTERN.match(plugin)
    version = None
    if match is not None:
        plugin = match.group(1)
        if match.group(2) is not None:
            version = match.group(2)

    for k, v in NAMESPACES.get_items():
        if v.plugin == plugin:
            yield k, Version(plugin, version)


def check_plugin_version(plugin: str, version: Version) -> bool:
    """Check if the plugin with given version is supported."""
    if plugin not in NAMESPACES:
        return False

    if version.version is None:  # Versionless plugin
        return True

    return version <= NAMESPACES.get_version(plugin)


class Namespace(dict):
    """Namespace dictionary to hold xml-element namespace data and versions."""

    def __getitem__(self, key: str) -> str:
        version = super().__getitem__(key)
        return "".join(
            attr for attr in version.values if isinstance(attr, str)
        )

    def items(self) -> cabc.ItemsView[str, str]:  # type: ignore[override]
        """Get original dictionary itemsview."""
        return {k: self.__getitem__(k) for k in super().keys()}.items()

    def get_items(self) -> cabc.ItemsView[str, Version]:
        """Get dictionary itemsview of name and version."""
        return super().items()

    def get_version(self, key: str) -> Version:
        """Get original namespace value."""
        return super().__getitem__(key)


#: These XML namespaces are defined in every ``.aird`` document
NAMESPACES = Namespace(
    {
        "CapellaRequirements": Version(
            "http://www.polarsys.org/capella/requirements"
        ),
        "Requirements": Version(
            "http://www.polarsys.org/kitalpha/requirements"
        ),
        "concern": Version(
            "http://www.eclipse.org/sirius/diagram/description/concern/",
            "1.1.0",
        ),
        "description": Version(
            "http://www.eclipse.org/sirius/description/", "1.1.0"
        ),
        "description_1": Version(
            "http://www.eclipse.org/sirius/diagram/description/", "1.1.0"
        ),
        "description_2": Version(
            "http://www.eclipse.org/sirius/table/description/", "1.1.0"
        ),
        "description_3": Version(
            "http://www.eclipse.org/sirius/diagram/sequence/description/",
            "2.0.0",
        ),
        "diagram": Version("http://www.eclipse.org/sirius/diagram/", "1.1.0"),
        "diagramstyler": Version("http://thalesgroup.com/mde/melody/ordering"),
        "ecore": Version("http://www.eclipse.org/emf/2002/Ecore"),
        "filter": Version(
            "http://www.eclipse.org/sirius/diagram/description/filter/",
            "1.1.0",
        ),
        "libraries": Version(
            "http://www.polarsys.org/capella/common/libraries/", "5.0.0"
        ),
        "metadata": Version(
            "http://www.polarsys.org/kitalpha/ad/metadata/", "1.0.0"
        ),
        "notation": Version(
            "http://www.eclipse.org/gmf/runtime/1.0.2/notation"
        ),
        "org.polarsys.capella.core.data.capellacommon": Version(
            "http://www.polarsys.org/capella/core/common/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.capellacore": Version(
            "http://www.polarsys.org/capella/core/core/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.capellamodeller": Version(
            "http://www.polarsys.org/capella/core/modeller/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.cs": Version(
            "http://www.polarsys.org/capella/core/cs/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.ctx": Version(
            "http://www.polarsys.org/capella/core/ctx/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.epbs": Version(
            "http://www.polarsys.org/capella/core/epbs/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.fa": Version(
            "http://www.polarsys.org/capella/core/fa/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.information": Version(
            "http://www.polarsys.org/capella/core/information/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.information.datatype": Version(
            "http://www.polarsys.org/capella/core/information/datatype/",
            "5.0.0",
        ),
        "org.polarsys.capella.core.data.information.datavalue": Version(
            "http://www.polarsys.org/capella/core/information/datavalue/",
            "5.0.0",
        ),
        "org.polarsys.capella.core.data.interaction": Version(
            "http://www.polarsys.org/capella/core/interaction/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.la": Version(
            "http://www.polarsys.org/capella/core/la/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.oa": Version(
            "http://www.polarsys.org/capella/core/oa/", "5.0.0"
        ),
        "org.polarsys.capella.core.data.pa": Version(
            "http://www.polarsys.org/capella/core/pa/", "5.0.0"
        ),
        "re": Version("http://www.polarsys.org/capella/common/re/", "1.3.0"),
        "sequence": Version(
            "http://www.eclipse.org/sirius/diagram/sequence/", "2.0.0"
        ),
        "style": Version(
            "http://www.eclipse.org/sirius/diagram/description/style/", "1.1.0"
        ),
        "table": Version("http://www.eclipse.org/sirius/table/", "1.1.0"),
        "viewpoint": Version("http://www.eclipse.org/sirius/", "1.1.0"),
        "xmi": Version("http://www.omg.org/XMI"),
        "xsi": Version("http://www.w3.org/2001/XMLSchema-instance"),
    }
)



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""The decoration factories for svg elements."""
from __future__ import annotations

import collections.abc as cabc
import logging
import re
import typing as t

logger = logging.getLogger(__name__)

icon_size = 20
"""Default icon size."""
icon_padding = 2
"""Default icon padding (left/right side)."""
feature_space = 24
"""Default margins/padding (top/bot and left/right) for feature text."""

function_ports = {"FIP", "FOP"}
directed_component_ports = {"CP_IN", "CP_OUT"}
component_ports = directed_component_ports | {"PP", "CP_INOUT", "CP_UNSET"}
all_ports = function_ports | component_ports
all_directed_ports = directed_component_ports | function_ports
start_aligned = {
    "LogicalComponent",
    "LogicalActor",
    "LogicalHumanActor",
    "LogicalHumanComponent",
}
only_icons = {"Requirement"}
DiagramClass = str
FaultyClass = str
PatchClass = str
always_top_label = {
    "Class",
    "Enumeration",
    "Note",
    "OperationalActivity",
    "PhysicalComponent",
}
needs_feature_line = {"Class", "Enumeration"}


class DecoFactories(t.Dict[str, cabc.Callable]):
    def __call__(self, func: cabc.Callable) -> cabc.Callable:
        symbol_name = re.sub(
            "(?:^|_)([a-z])",
            lambda m: m.group(1).capitalize(),
            func.__name__,
        )
        self[symbol_name] = func
        return func

    def __missing__(self, class_: str) -> cabc.Callable:
        logger.error("%s wasn't found in factories.", class_)
        assert "ErrorSymbol" in self
        return self["ErrorSymbol"]


deco_factories = DecoFactories()



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Core functionality shared by all PVMT submodules."""

from capellambse.loader.xmltools import AttributeProperty
from capellambse.loader.xmltools import XMLDictProxy as XMLDictProxy_

NAMESPACES = dict(xsi="http://www.w3.org/2001/XMLSchema-instance")


class Generic:
    """Base class for PropertyValues, Domains etc."""

    idx = AttributeProperty("xml_element", "id", writable=False)
    # TODO Investigate potential problems with renaming things
    name = AttributeProperty("xml_element", "name")
    sid = AttributeProperty("xml_element", "sid", optional=True)

    def __init__(self, xml_element, *args, **kwargs):
        super().__init__(*args, **kwargs)  # type: ignore[call-arg]
        self.xml_element = xml_element

    def __repr__(self):
        return '<pvmt.{} "{}"({})>'.format(
            type(self).__name__, self.name, self.idx
        )

    @classmethod
    def from_xml_element(cls, element):
        """Construct an object from the given XML element.

        This function is used to allow subclasses more control over how
        they are instantiated from existing XML elements, compared to
        creating them from scratch.
        """
        return cls(element)


class XMLDictProxy(Generic, XMLDictProxy_):  # pylint: disable=abstract-method
    """Facilitates usage of the :class:`capellambse.loader.XMLDictProxy`."""

    def __init__(self, xml_element, *args, **kwargs):
        # pylint: disable=arguments-out-of-order
        super().__init__(xml_element, xml_element, *args, **kwargs)



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""The MelodyLoader loads and provides access to a Capella model.

It is using LXML internally to efficiently parse and navigate through
the Capella-generated XML files.  For more information about LXML, see
the `LXML Documentation`_.

.. _LXML Documentation: https://lxml.de/
"""

from .core import *
from .filehandler import FileHandler, get_filehandler
from .modelinfo import ModelInfo



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from .. import common as c


@c.xtype_handler(None)
class Constraint(c.GenericElement):
    """A constraint."""

    _xmltag = "ownedConstraints"

    constrained_elements = c.AttrProxyAccessor(
        c.GenericElement,
        "constrainedElements",
        aslist=c.MixedElementList,
    )

    specification = c.SpecificationAccessor()


@c.xtype_handler(None)
class Generalization(c.GenericElement):
    """A Generalization"""

    _xmltag = "ownedGeneralizations"


c.set_accessor(
    c.GenericElement,
    "constraints",
    c.ProxyAccessor(Constraint, aslist=c.ElementList),
)



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""The AIRD parser and various other diagramming related tools.

This module is used to enumerate, access and export the diagrams from
Capella projects.  The JSON output it produces can be fed for example
into the :mod:`capellambse.svg` module for conversion to SVG.
"""
# isort: off
from .vector2d import *

from .capstyle import *
from .diagram import *
from .json_enc import *
from .parser import *



#!/usr/bin/env python
# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys

if __name__ == "__main__":
    if "username" in sys.argv[1].lower():
        print(os.environ["GIT_USERNAME"])
    elif "password" in sys.argv[1].lower():
        print(os.environ["GIT_PASSWORD"])
    else:
        raise SystemExit(1)



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Implementation of objects and relations for Information capture and data modelling

Information objects inheritance tree (taxonomy):

.. diagram:: [CDB] Information [Taxonomy]

Information object-relations map (ontology):

.. diagram:: [CDB] Information [Ontology]
"""

from __future__ import annotations

from capellambse.loader import xmltools

from ... import common as c
from ... import modeltypes
from .. import capellacommon, capellacore
from . import datatype, datavalue


def _allocated_exchange_items(
    obj: c.GenericElement,
) -> c.ElementList[c.GenericElement]:
    try:
        return obj.exchange_items
    except AttributeError:
        pass

    try:
        return obj.allocated_exchange_items
    except AttributeError:
        pass

    raise TypeError(
        f"Unhandled exchange type: {type(obj).__name__}"
    )  # pragma: no cover


def _search_all_exchanges(
    obj: c.GenericElement,
) -> c.ElementList[c.GenericElement]:
    from .. import fa

    return obj._model.search(fa.ComponentExchange, fa.FunctionalExchange)


@c.xtype_handler(None)
class Unit(c.GenericElement):
    """Unit"""

    _xmltag = "ownedUnits"


@c.xtype_handler(None)
class Property(c.GenericElement):
    """A Property of a Class."""

    _xmltag = "ownedFeatures"

    is_ordered = xmltools.BooleanAttributeProperty(
        "_element",
        "ordered",
        __doc__="Boolean flag, indicates if property is ordered",
    )
    is_unique = xmltools.BooleanAttributeProperty(
        "_element",
        "unique",
        __doc__="Boolean flag, indicates if property is unique",
    )
    is_abstract = xmltools.BooleanAttributeProperty(
        "_element",
        "isAbstract",
        __doc__="Boolean flag, indicates if property is abstract",
    )
    is_static = xmltools.BooleanAttributeProperty(
        "_element",
        "isStatic",
        __doc__="Boolean flag, indicates if property is static",
    )
    is_part_of_key = xmltools.BooleanAttributeProperty(
        "_element",
        "isPartOfKey",
        __doc__="Boolean flag, indicates if property is part of key",
    )
    is_derived = xmltools.BooleanAttributeProperty(
        "_element",
        "isDerived",
        __doc__="Boolean flag, indicates if property is abstract",
    )
    is_read_only = xmltools.BooleanAttributeProperty(
        "_element",
        "isReadOnly",
        __doc__="Boolean flag, indicates if property is read-only",
    )
    visibility = xmltools.EnumAttributeProperty(
        "_element", "visibility", modeltypes.VisibilityKind, default="UNSET"
    )
    type = c.AttrProxyAccessor(c.GenericElement, "abstractType")
    default_value = c.RoleTagAccessor("ownedDefaultValue")
    min = c.RoleTagAccessor("ownedMinValue")
    max = c.RoleTagAccessor("ownedMaxValue")
    null_value = c.RoleTagAccessor("ownedNullValue")
    min_card = c.RoleTagAccessor("ownedMinCard")
    max_card = c.RoleTagAccessor("ownedMaxCard")


@c.xtype_handler(None)
class Class(c.GenericElement):
    """A Class."""

    _xmltag = "ownedClasses"

    sub: c.Accessor
    super: c.Accessor[Class]
    is_abstract = xmltools.BooleanAttributeProperty(
        "_element",
        "abstract",
        __doc__="Boolean flag, indicates if class is abstract",
    )
    is_final = xmltools.BooleanAttributeProperty(
        "_element",
        "final",
        __doc__="Boolean flag, indicates if class is final",
    )
    is_primitive = xmltools.BooleanAttributeProperty(
        "_element",
        "isPrimitive",
        __doc__="Boolean flag, indicates if class is primitive",
    )
    visibility = xmltools.EnumAttributeProperty(
        "_element", "visibility", modeltypes.VisibilityKind, default="UNSET"
    )
    state_machines = c.ProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )
    owned_properties = c.ProxyAccessor(
        Property, aslist=c.ElementList, follow_abstract=False
    )

    @property
    def properties(self) -> c.ElementList[Property]:
        """Return all owned and inherited properties."""
        return (
            self.owned_properties + self.super.properties
            if self.super is not None
            else self.owned_properties
        )


@c.xtype_handler(None)
class Union(Class):
    """A Union."""

    _xmltag = "ownedClasses"

    kind = xmltools.EnumAttributeProperty(
        "_element", "kind", modeltypes.UnionKind, default="UNION"
    )


@c.xtype_handler(None)
class Collection(c.GenericElement):
    """A Collection."""

    _xmltag = "ownedCollections"

    kind = xmltools.EnumAttributeProperty(
        "_element", "kind", modeltypes.CollectionKind, default="ARRAY"
    )

    sub: c.Accessor
    super: c.Accessor[Collection]


@c.xtype_handler(None)
class DataPkg(c.GenericElement):
    """A data package that can hold classes."""

    classes = c.ProxyAccessor(Class, aslist=c.ElementList)
    unions = c.ProxyAccessor(Union, aslist=c.ElementList)
    collections = c.ProxyAccessor(Collection, aslist=c.ElementList)
    enumerations = c.ProxyAccessor(datatype.Enumeration, aslist=c.ElementList)
    complex_values = c.ProxyAccessor(
        datavalue.ComplexValue, aslist=c.ElementList, follow_abstract=False
    )
    packages: c.Accessor


@c.xtype_handler(None)
class ExchangeItemElement(c.GenericElement):
    """An ExchangeItemElement (proxy link)."""

    _xmltag = "ownedElements"

    abstract_type = c.AttrProxyAccessor(c.GenericElement, "abstractType")
    owner: c.Accessor


@c.xtype_handler(None)
class ExchangeItem(c.GenericElement):
    """An item that can be exchanged on an Exchange."""

    _xmltag = "ownedExchangeItems"

    type = xmltools.EnumAttributeProperty(
        "_element",
        "exchangeMechanism",
        modeltypes.ExchangeItemType,
        default="UNSET",
    )
    elements = c.ProxyAccessor(
        ExchangeItemElement,
        aslist=c.ElementList,
        follow_abstract=False,
    )
    exchanges = c.CustomAccessor(
        c.GenericElement,
        _search_all_exchanges,
        matchtransform=_allocated_exchange_items,
        aslist=c.ElementList,
    )


for cls in [Class, Union, datatype.Enumeration, Collection]:
    c.set_accessor(
        cls,
        "super",
        c.ProxyAccessor(
            cls,
            capellacore.Generalization,
            follow="super",
            follow_abstract=False,
        ),
    )
    c.set_accessor(
        cls,
        "sub",
        c.ReferenceSearchingAccessor(cls, "super", aslist=c.MixedElementList),
    )

c.set_accessor(
    datavalue.EnumerationLiteral,
    "owner",
    c.ParentAccessor(datatype.Enumeration),
)
c.set_accessor(
    DataPkg, "packages", c.ProxyAccessor(DataPkg, aslist=c.ElementList)
)
c.set_accessor(ExchangeItemElement, "owner", c.ParentAccessor(ExchangeItem))



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from .. import common as c

XT_CAP2PROC = "org.polarsys.capella.core.data.interaction:FunctionalChainAbstractCapabilityInvolvement"
XT_CAP2ACT = "org.polarsys.capella.core.data.interaction:AbstractFunctionAbstractCapabilityInvolvement"
XT_CAP_REAL = (
    "org.polarsys.capella.core.data.interaction:AbstractCapabilityRealization"
)


@c.xtype_handler(None)
class Scenario(c.GenericElement):
    """A scenario that holds instance roles."""


class Exchange(c.GenericElement):
    """An abstract Exchange."""

    source = c.ParentAccessor(c.GenericElement)


@c.xtype_handler(None)
class AbstractCapabilityExtend(Exchange):
    """An AbstractCapabilityExtend."""

    _xmltag = "extends"

    target = c.AttrProxyAccessor(c.GenericElement, "extended")


@c.xtype_handler(None)
class AbstractCapabilityInclude(Exchange):
    """An AbstractCapabilityInclude."""

    _xmltag = "includes"

    target = c.AttrProxyAccessor(c.GenericElement, "included")


@c.xtype_handler(None)
class AbstractCapabilityGeneralization(Exchange):
    """An AbstractCapabilityGeneralization."""

    _xmltag = "superGeneralizations"

    target = c.AttrProxyAccessor(c.GenericElement, "super")


class AbstractInvolvement(c.GenericElement):
    """An abstract Involvement."""

    involved = c.AttrProxyAccessor(c.GenericElement, "involved")

    @property
    def name(self) -> str:  # type: ignore
        return f"[{self.__class__.__name__}] to {self.involved.name} ({self.involved.uuid})"


@c.xtype_handler(None)
class AbstractFunctionAbstractCapabilityInvolvement(AbstractInvolvement):
    """An abstract CapabilityInvolvement linking to SystemFunctions."""



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Custom extensions to the svgwrite ``Drawing`` object."""
from __future__ import annotations

import collections.abc as cabc
import dataclasses
import logging
import os
import re
import typing as t

from svgwrite import base, container, drawing, shapes
from svgwrite import text as svgtext

from capellambse import aird
from capellambse import helpers as chelpers

from . import decorations, generate, helpers, style, symbols

LOGGER = logging.getLogger(__name__)
DEBUG = "CAPELLAMBSE_SVG_DEBUG" in os.environ
"""Debug flag to render helping lines."""
LABEL_ICON_PADDING = 2
"""Default padding between a label's icon and text."""


LabelDict = t.TypedDict(
    "LabelDict",
    {
        "x": float,
        "y": float,
        "width": float,
        "height": float,
        "text": str,
        "class": str,
    },
    total=False,
)


# FIXME: refactor this, so a Drawing contains an "svg drawing". Always prefer composition over inheritance.
class Drawing(drawing.Drawing):
    """The main container that stores all svg elements."""

    def __init__(self, metadata: generate.DiagramMetadata):
        superparams = {
            "filename": f"{metadata.name}.svg",
            "size": metadata.size,
            "viewBox": metadata.viewbox,
        }
        if metadata.class_:
            superparams["class_"] = re.sub(r"\s+", "", metadata.class_)

        super().__init__(**superparams)
        self.diagram_class = metadata.class_
        self.stylesheet = self.make_stylesheet()
        self.add_backdrop(pos=metadata.pos, size=metadata.size)

        self.obj_cache: dict[str | None, t.Any] = {}

    def add_backdrop(
        self, pos: tuple[float, float], size: tuple[float, float]
    ) -> None:
        """Add a white background rectangle to the drawing."""
        self.__backdrop = self.rect(
            insert=pos,
            size=size,
            fill="white",
            stroke="none",
        )
        self.add(self.__backdrop)

    def make_stylesheet(self) -> style.SVGStylesheet:
        """Return created stylesheet and add sheet and decorations to defs."""
        stylesheet = style.SVGStylesheet(class_=self.diagram_class or "")
        self.defs.add(stylesheet.sheet)
        for name in stylesheet.static_deco:
            self.defs.add(decorations.deco_factories[name]())

        for grad in stylesheet.yield_gradients():
            self.defs.add(grad)

        return stylesheet

    def __repr__(self) -> str:
        return super()._repr_svg_()

    def add_rect(
        self,
        pos: tuple[float, float],
        size: tuple[float, float],
        rectstyle: cabc.Mapping[str, style.Styling],
        *,
        class_: str = "",
        label: LabelDict | None = None,
        features: cabc.Sequence[str] = (),
        id_: str | None = None,
        children: bool = False,
    ) -> container.Group:
        """Add a rectangle with auto-aligned text to the group."""
        grp: container.Group = self.g(class_=f"Box {class_}", id_=id_)
        text_style = rectstyle["text_style"]
        rect_style = rectstyle["obj_style"]
        transform = chelpers.get_transformation(class_, pos, size)
        rectparams: dict[str, t.Any] = {
            "insert": pos,
            "size": size,
            "class_": class_,
            **transform,
        }
        if rect_style:
            rectparams["style"] = rect_style[""]

        rect: shapes.Rect = self.rect(**rectparams)
        grp.add(rect)

        if features or class_ in decorations.needs_feature_line:
            self._draw_feature_line(rect, grp, rect_style)
            if features:
                self._draw_feature_text(
                    rect, features, class_, grp, text_style
                )

        if label:
            text_anchor = (
                "start" if class_ in decorations.start_aligned else "middle"
            )
            y_margin = None
            if children or class_ in decorations.always_top_label:
                y_margin = 5

            self._draw_box_label(
                LabelBuilder(
                    label,
                    grp,
                    labelstyle=text_style,
                    class_=class_,
                    text_anchor=text_anchor,
                    y_margin=y_margin,
                )
            )
        elif class_ in decorations.only_icons:
            icon_size = 50
            labelpos = (
                pos[0] + (size[0] - icon_size / 4) / 2,
                pos[1] + (decorations.feature_space - 15) / 2,
            )
            label = {  # XXX: Instead of allowing None > more asserts
                "x": labelpos[0],
                "y": labelpos[1],
                "width": 0.0,
                "height": 0.0,
                "text": "",
                "class": "",
            }
            self.add_label_image(
                LabelBuilder(
                    label, grp, text_style, class_=class_, icon_size=icon_size
                ),
                labelpos,
            )

        if DEBUG:
            helping_lines = self._draw_rect_helping_lines(pos, size)
            for line in helping_lines:
                grp.add(line)

        return grp

    def _draw_feature_line(
        self,
        obj: base.BaseElement,
        group: container.Group | None,
        objstyle: style.Styling | None,
    ) -> shapes.Line | None:
        """Draw a Line on the given object."""
        x, y = obj.attribs["x"], obj.attribs["y"]
        w = obj.attribs["width"]
        css: str | None = None
        if objstyle is not None:
            css = objstyle["stroke"] if "stroke" in objstyle else None

        line = self.line(
            start=(x, y + decorations.feature_space),
            end=(x + w, y + decorations.feature_space),
            style=css,
        )
        if group is None:
            return line

        return group.add(line)

    def _draw_feature_text(
        self,
        obj: base.BaseElement,
        features: cabc.Sequence[str],
        class_: str,
        group: container.Group,
        labelstyle: style.Styling,
    ) -> None:
        """Draw features text on given object."""
        x, y = obj.attribs["x"], obj.attribs["y"]
        w, h = obj.attribs["width"], obj.attribs["height"]
        label: LabelDict = {
            "x": x,
            "y": y + decorations.feature_space,
            "width": w,
            "height": h - decorations.feature_space,
            "class": "Features",
            "text": "\n".join(
                chelpers.flatten_html_string(feat) for feat in features
            ),
        }
        self._draw_box_label(
            LabelBuilder(
                label,
                group,
                class_=class_,
                labelstyle=labelstyle,
                y_margin=7,
                icon=False,
            )
        )

    def _draw_box_label(self, builder: LabelBuilder) -> container.Group:
        """Draw label text on given object and return calculated label position."""
        x, text_height, _, y_margin = self._draw_label(builder)

        if DEBUG:
            assert builder.label is not None
            assert y_margin is not None
            debug_y = builder.label["y"] + y_margin
            debug_y1 = (
                builder.label["y"]
                + (builder.label["height"] - decorations.icon_size) / 2
            )
            x = (
                builder.label["x"]
                + decorations.icon_size
                + 2 * decorations.icon_padding
            )
            if text_height >= decorations.icon_size:
                debug_height = text_height
            else:
                debug_height = decorations.icon_size

            bbox: LabelDict = {
                "x": x
                if builder.text_anchor == "start"
                else x - builder.label["width"] / 2,
                "y": debug_y if debug_y <= debug_y1 else debug_y1,
                "width": builder.label["width"],
                "height": debug_height,
            }
            labelstyle = style.Styling(
                self.diagram_class,
                "Box",
                stroke="rgb(239, 41, 41)",
                fill="none",
            )
            self._draw_label_bbox(
                bbox, builder.group, "Box", obj_style=labelstyle
            )
            self._draw_circle(
                center_=(builder.label["x"], builder.label["y"]),
                radius_=3,
                obj_style=style.Styling(
                    self.diagram_class,
                    "Box",
                    fill="rgb(239, 41, 41)",
                ),
                text_style=None,
            )

        return builder.group

    def _draw_label(
        self, builder: LabelBuilder
    ) -> tuple[float, float, float, float | int | None]:
        assert isinstance(builder.label["x"], (int, float))
        assert isinstance(builder.label["y"], (int, float))
        assert isinstance(builder.label["width"], (int, float))
        assert isinstance(builder.label["height"], (int, float))
        assert "text" not in builder.label or isinstance(
            builder.label["text"], str
        )
        text = self.text(
            text="",
            insert=(builder.label["x"], builder.label["y"]),
            class_=builder.label["class"],
            text_anchor=builder.text_anchor,
            dominant_baseline="middle",
            style=builder.labelstyle[""],
        )
        builder.group.add(text)

        render_icon = (
            f"{builder.class_}Symbol" in decorations.deco_factories
            and builder.icon
        )
        lines = render_hbounded_lines(builder, render_icon)
        x, icon_x = get_label_position_x(builder, lines, render_icon)
        y = get_label_position_y(builder, lines)
        for line in lines.lines:
            text.add(
                svgtext.TSpan(
                    insert=(x, y), text=line, **{"xml:space": "preserve"}
                )
            )
            y += lines.line_height

        if render_icon:
            icon_pos = get_label_icon_position(
                builder, lines.text_height, icon_x
            )
            self.add_label_image(builder, icon_pos)

        return x, lines.text_height, lines.max_line_width, builder.y_margin

    def add_label_image(
        self, builder: LabelBuilder, pos: tuple[float, float]
    ) -> None:
        """Add label svgobject to given group."""
        if builder.class_ is None:
            builder.class_ = "Error"

        builder.group.add(
            self.use(
                href=f"#{builder.class_}Symbol",
                insert=pos,
                size=(builder.icon_size, builder.icon_size),
            )
        )

    def get_port_transformation(
        self,
        pos: tuple[float, float],
        size: tuple[float, float],
        class_: str,
        parent_id: str | None,
    ) -> str:
        """Return rotation transformation for port-object-styling."""
        rect = self.obj_cache[parent_id]
        x, y = pos
        w, h = size
        angle = 0
        par_x, par_y = rect["x"] + 0.5, rect["y"] + 0.5
        if x <= par_x:
            angle = 90
        elif par_x + rect["width"] < x + w:
            angle = -90
        elif y <= par_y:
            angle = -180
        angle += 180 * (class_ in ["FOP", "CP_IN"])

        return "rotate({} {} {})".format(angle, x + w / 2, y + h / 2)

    def add_port(
        self,
        pos: tuple[float, float],
        size: tuple[float, float],
        text_style: style.Styling,
        parent_id: str | None,
        *,
        class_: str,
        obj_style: style.Styling | None = None,
        label: LabelDict | None = None,
        id_: str | None = None,
    ) -> container.Group:
        styles = None if obj_style is None else obj_style[""]
        grp = self.g(class_=f"Box {class_}", id_=id_)
        if class_ in decorations.all_directed_ports:
            port_id = "#ErrorSymbol"
            if class_ in decorations.function_ports:
                port_id = "#PortSymbol"
            elif class_ in decorations.component_ports:
                port_id = "#ComponentPortSymbol"

            grp.style = styles
            grp.add(
                self.use(
                    href=port_id,
                    insert=pos,
                    size=size,
                    transform=self.get_port_transformation(
                        pos, size, class_, parent_id
                    ),
                    class_=class_,
                )
            )
        else:
            grp.add(
                self.rect(
                    insert=pos,
                    size=size,
                    class_=class_,
                    transform=self.get_port_transformation(
                        pos, size, class_, parent_id
                    ),
                    style=styles,
                )
            )

        if label is not None:
            self._draw_label(
                LabelBuilder(
                    label,
                    grp,
                    class_="Annotation",
                    labelstyle=text_style,
                    text_anchor="middle",
                    y_margin=0,
                    icon=False,
                )
            )

        return grp

    def draw_object(self, obj: cabc.Mapping[str, t.Any]) -> None:
        """Draw an object into this drawing.

        Parameters
        ----------
        obj
            The (decoded) JSON-dict of a single diagram object.
        """
        try:
            drawfunc: t.Any = getattr(self, f'_draw_{obj["type"]}')
        except AttributeError:
            raise ValueError(f'Invalid object type: {obj["type"]}') from None

        objparams = {
            f"{k}_": v for k, v in obj.items() if k not in {"type", "style"}
        }
        if obj["class"] in decorations.all_ports:
            obj["type"] = "box"  # type: ignore[index]  # FIXME: *HACK* should actually copy the object before modifying

        class_: str = obj["type"].capitalize() + (
            f".{obj['class']}" if "class" in obj else ""
        )
        obj_style = style.Styling(
            self.diagram_class,
            class_,
            **{k: v for k, v in obj.get("style", {}).items() if "_" not in k},
        )
        text_style = style.Styling(
            self.diagram_class,
            class_,
            "text",
            **{
                k[len("text_") :]: v
                for k, v in obj.get("style", {}).items()
                if k.startswith("text_")
            },
        )

        self.obj_cache[obj["id"]] = obj

        drawfunc(**objparams, obj_style=obj_style, text_style=text_style)

        self._deploy_defs(obj_style)
        self._deploy_defs(text_style)

    def _deploy_defs(self, styling: style.Styling) -> None:
        defs_ids = {d.attribs.get("id") for d in self.defs.elements}
        for attr in styling:
            val = getattr(styling, attr)
            if isinstance(val, cabc.Iterable) and not isinstance(
                val, (str, aird.RGB)
            ):
                grad_id = styling._generate_id("CustomGradient", val)
                if grad_id not in defs_ids:
                    gradient = symbols._make_lgradient(
                        id_=grad_id, stop_colors=val
                    )
                    self.defs.add(gradient)
                    defs_ids.add(grad_id)

        defaultstyles = aird.get_style(self.diagram_class, styling._class)

        def getstyleattr(sobj: object, attr: str) -> t.Any:
            return getattr(sobj, attr, None) or defaultstyles.get(
                styling._style_name(attr)
            )

        markers = (
            getstyleattr(super(), "marker-start"),
            getstyleattr(super(), "marker-end"),
        )
        for marker in markers:
            if marker is None:
                continue

            stroke = str(getstyleattr(styling, "stroke"))
            stroke_width = str(getstyleattr(styling, "stroke-width"))
            marker_id = styling._generate_id(marker, [stroke])
            if marker_id not in defs_ids:
                self.defs.add(
                    decorations.deco_factories[marker](
                        marker_id,
                        style=style.Styling(
                            self.diagram_class,
                            styling._class,
                            _prefix=styling._prefix,
                            fill=stroke,
                            stroke=stroke,
                            stroke_width=stroke_width,
                        ),
                    )
                )
                defs_ids.add(marker_id)

    def _draw_symbol(
        self,
        *,
        x_: int | float,
        y_: int | float,
        width_: int | float,
        height_: int | float,
        context_: cabc.Sequence[str] = (),
        parent_: str | None = None,
        label_: LabelDict | None = None,
        id_: str | None = None,
        class_: str,
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw: t.Any,
    ):
        del obj_style, context_  # FIXME Add context to SVG
        del kw  # Dismiss additional info from json
        assert isinstance(label_, (dict, type(None)))
        pos = (x_ + 0.5, y_ + 0.5)
        size = (width_, height_)

        if class_ in decorations.all_ports:
            grp = self.add_port(
                pos,
                size,
                text_style,
                parent_,
                class_=class_,
                label=label_,
                id_=id_,
            )
        else:
            grp = self.g(class_=f"Box {class_}", id_=id_)
            grp.add(
                self.use(
                    href=f"#{class_}Symbol",
                    insert=pos,
                    size=size,
                    class_=class_,
                )
            )

        self.add(grp)
        if label_:
            label_["class"] = "Annotation"
            self._draw_label(
                LabelBuilder(
                    label_,
                    grp or self.elements[-1],
                    labelstyle=text_style,
                    class_=class_,
                    y_margin=0,
                    text_anchor="middle",
                    icon=False,
                )
            )

    def _draw_box(
        self,
        *,
        x_: int | float,
        y_: int | float,
        width_: int | float,
        height_: int | float,
        context_: cabc.Sequence[str] = (),
        children_: cabc.Sequence[str] = (),
        features_: cabc.Sequence[str] = (),
        label_: str | LabelDict | None = None,
        id_: str,
        class_: str,
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw: t.Any,
    ) -> container.Group:
        del context_  # FIXME Add context to SVG
        del kw  # Dismiss additional info from json, for e.g.: ports_
        pos = (x_ + 0.5, y_ + 0.5)
        size = (width_, height_)
        rect_style = {"text_style": text_style, "obj_style": obj_style}
        label: LabelDict | None = None
        if isinstance(label_, str):
            label = {
                "x": x_,
                "y": y_,
                "width": width_,
                "height": height_,
                "text": label_,
                "class": class_ or "Box",
            }
        elif label_ is not None:
            label = label_
            label["class"] = "Annotation"

        grp = self.add_rect(
            pos,
            size,
            rect_style,
            class_=class_,
            id_=id_,
            label=label,
            features=features_,
            children=bool(children_),
        )
        return self.add(grp)

    def _draw_box_symbol(
        self,
        *,
        x_: int | float,
        y_: int | float,
        width_: int | float,
        height_: int | float,
        label_: LabelDict,
        id_: str,
        class_: str,
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw: t.Any,
    ) -> None:
        del kw

        grp = self._draw_box(
            x_=x_,
            y_=y_,
            width_=width_,
            height_=height_,
            id_=id_,
            class_=class_,
            obj_style=obj_style,
            text_style=text_style,
        )
        label_["class"] = "Annotation"
        self._draw_label(
            LabelBuilder(
                label_,
                grp or self.elements[-1],
                labelstyle=text_style,
                class_=class_,
                y_margin=0,
                text_anchor="middle",
                icon=False,
            )
        )

    def _draw_circle(
        self, *, center_, radius_, id_=None, class_=None, obj_style, text_style
    ):
        del text_style  # No label for circles
        center_ = tuple(i + 0.5 for i in center_)

        grp = self.g(class_=f"Circle {class_}", id_=id_)
        grp.add(self.circle(center=center_, r=radius_, style=obj_style[""]))
        self.add(grp)

    def _draw_edge(
        self,
        *,
        points_: list[list[int]],
        labels_: t.Sequence[LabelDict] = (),
        id_: str,
        class_: str,
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw,
    ):
        del kw  # Dismiss additional info from json
        points: list = [(x + 0.5, y + 0.5) for x, y in points_]
        grp = self.g(class_=f"Edge {class_}", id_=id_)
        grp.add(
            self.path(d=["M"] + points, class_="Edge", style=obj_style[""])
        )

        # Received text space doesn't allow for anything else than the text
        for label_ in labels_:
            label_["class"] = "Annotation"
            self._draw_label_bbox(label_, grp, "AnnotationBB")
            self._draw_edge_label(
                label_,
                grp,
                labelstyle=text_style,
                text_anchor="middle",
                class_=class_,
                y_margin=0,
            )

        self.add(grp)

    def _draw_edge_label(
        self,
        label: LabelDict,
        group: container.Group,
        *,
        class_: str,
        labelstyle: style.Styling,
        text_anchor: str = "start",
        y_margin: int | float,
    ) -> container.Group:
        class_ = (
            style.get_symbol_styleclass(class_, self.diagram_class) or class_
        )
        if f"{class_}Symbol" in decorations.deco_factories:
            additional_space = (
                decorations.icon_size + 2 * decorations.icon_padding
            )
            label["width"] += additional_space + 2
            label["x"] -= additional_space / 2

        self._draw_label(
            LabelBuilder(
                label,
                group,
                labelstyle=labelstyle,
                class_=class_,
                y_margin=y_margin,
                text_anchor=text_anchor,
            )
        )

        return group

    def _draw_label_bbox(
        self,
        label: LabelDict,
        group: container.Group | None = None,
        class_: str | None = None,
        obj_style: style.Styling | None = None,
    ) -> None:
        """Draw a bounding box for given label."""
        if DEBUG:
            if obj_style is not None:
                setattr(obj_style, "stroke", "rgb(239, 41, 41);")
            else:
                obj_style = style.Styling(
                    self.diagram_class,
                    "AnnotationBB",
                    stroke="rgb(239, 41, 41);",
                )

        bbox = self.rect(
            insert=(label["x"], label["y"]),
            size=(label["width"], label["height"]),
            class_=class_,
            style=obj_style,
        )
        if group is None:
            return bbox

        return group.add(bbox)

    def _draw_line(
        self,
        data: dict[str, float | int],
        group: container.Group | None = None,
        obj_style: style.Styling | None = None,
    ) -> shapes.Line | None:
        """Draw a Line on the given object."""
        x1, y1 = data["x"], data["y"]
        x2 = data.get("x1") or data["x"] + data["width"]
        y2 = data.get("y1") or data["y"]
        line = self.line(start=(x1, y1), end=(x2, y2), style=obj_style)
        if group is None:
            return line

        return group.add(line)

    def _draw_rect_helping_lines(
        self,
        rect_pos: tuple[float, float],
        rect_size: tuple[float, float],
    ):
        linestyle = style.Styling(
            self.diagram_class,
            "Edge",
            **{
                "stroke": "rgb(239, 41, 41)",
                "stroke-dasharray": "5",
            },
        )
        lines = []
        lines.append(
            self._draw_line(
                {
                    "x": rect_pos[0] + rect_size[0] / 2,
                    "y": rect_pos[1],
                    "x1": rect_pos[0] + rect_size[0] / 2,
                    "y1": rect_pos[1] + rect_size[1],
                },
                obj_style=linestyle,
            )
        )
        lines.append(
            self._draw_line(
                {
                    "x": rect_pos[0],
                    "y": rect_pos[1] + rect_size[1] / 2,
                    "width": rect_size[0],
                },
                obj_style=linestyle,
            )
        )
        return lines


@dataclasses.dataclass
class LabelBuilder:
    label: LabelDict
    group: container.Group
    labelstyle: style.Styling
    class_: str | None = None
    y_margin: int | float | None = None
    text_anchor: str = "start"
    icon: bool = True
    icon_size: float | int = decorations.icon_size


class LinesData(t.NamedTuple):
    lines: list[str]
    line_height: float
    text_height: float
    margin: float
    max_line_width: float


def render_hbounded_lines(
    builder: LabelBuilder, render_icon: bool
) -> LinesData:
    (
        lines,
        label_margin,
        max_text_width,
    ) = helpers.check_for_horizontal_overflow(
        str(builder.label["text"]),
        builder.label["width"],
        decorations.icon_padding if render_icon else 0,
        builder.icon_size if render_icon else 0,
    )
    lines_to_render = helpers.check_for_vertical_overflow(
        lines, builder.label["height"], max_text_width
    )
    line_height = max(j for _, j in map(chelpers.extent_func, lines_to_render))
    text_height = line_height * len(lines_to_render)
    max_line_width = max(
        w for w, _ in map(chelpers.extent_func, lines_to_render)
    )
    assert max_text_width >= max_line_width
    return LinesData(
        lines_to_render, line_height, text_height, label_margin, max_line_width
    )


def get_label_position_x(
    builder: LabelBuilder, lines: LinesData, render_icon: bool
) -> tuple[float, float]:
    """Return x-coordinate of label-text and icon."""
    if builder.text_anchor == "start":
        x = (
            builder.label["x"]
            + lines.margin
            + (builder.icon_size + decorations.icon_padding) * render_icon
        )
        return x, builder.label["x"] + lines.margin
    x = builder.label["x"] + builder.label["width"] / 2
    return x, x - lines.max_line_width / 2 - builder.icon_size


def get_label_position_y(builder: LabelBuilder, lines: LinesData) -> float:
    """Return y-coordinate of label-text."""
    dominant_baseline_adjust = chelpers.extent_func(lines.lines[0])[1] / 2
    if builder.y_margin is None:
        builder.y_margin = (builder.label["height"] - lines.text_height) / 2
    return builder.label["y"] + builder.y_margin + dominant_baseline_adjust


def get_label_icon_position(
    builder: LabelBuilder, text_height: int | float, icon_x: int | float
) -> tuple[float, float]:
    """Calculate the position of the icon relative to the label.

    Parameters
    ----------
    builder : LabelBuilder
        LabelBuilder
    text_height : int | float
        The height of the text in the label.
    icon_x : int | float
        The x position of the icon.

    Returns
    -------
    icon_coords : tuple[float, float]
        The x and y coordinates of the icon.
    """
    assert builder.y_margin is not None
    icon_y = (
        builder.label["y"]
        + builder.y_margin
        + (text_height - builder.icon_size) / 2
    )
    if (
        icon_x < builder.label["x"] - decorations.icon_padding
        and builder.label["class"] != "Annotation"
    ):
        icon_x = builder.label["x"] - decorations.icon_padding
    return icon_x, icon_y



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Tools for the Logical Architecture layer.

.. diagram:: [CDB] LA ORM
"""
from __future__ import annotations

import operator

from .. import common as c
from .. import crosslayer, diagram
from ..crosslayer import capellacommon, capellacore, cs, fa, interaction
from . import ctx

XT_ARCH = "org.polarsys.capella.core.data.la:LogicalArchitecture"


@c.xtype_handler(XT_ARCH)
class LogicalFunction(fa.Function):
    """A logical function on the Logical Architecture layer."""

    _xmltag = "ownedLogicalFunctions"

    realized_system_functions = c.ProxyAccessor(
        ctx.SystemFunction,
        fa.FunctionRealization,
        aslist=c.ElementList,
        follow="targetElement",
    )

    @property
    def owner(self) -> LogicalComponent | None:
        try:
            return next(
                i
                for i in self._model.search("LogicalComponent")
                if self in i.functions
            )
        except StopIteration:
            return None


@c.xtype_handler(XT_ARCH)
class LogicalFunctionPkg(c.GenericElement):
    """A logical function package."""

    _xmltag = "ownedFunctionPkg"

    functions = c.ProxyAccessor(LogicalFunction, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class LogicalComponent(cs.Component):
    """A logical component on the Logical Architecture layer."""

    _xmltag = "ownedLogicalComponents"

    functions = c.ProxyAccessor(
        LogicalFunction,
        fa.XT_FCALLOC,
        aslist=c.ElementList,
        follow="targetElement",
    )
    realized_components = c.ProxyAccessor(
        ctx.SystemComponent,
        cs.ComponentRealization,
        aslist=c.ElementList,
        follow="targetElement",
    )

    components: c.Accessor


@c.xtype_handler(XT_ARCH)
class LogicalComponentPkg(c.GenericElement):
    """A logical component package."""

    _xmltag = "ownedLogicalComponentPkg"

    components = c.ProxyAccessor(LogicalComponent, aslist=c.ElementList)
    state_machines = c.ProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CapabilityRealization(c.GenericElement):
    """A capability."""

    _xmltag = "ownedCapabilityRealizations"

    owned_chains = c.ProxyAccessor(fa.FunctionalChain, aslist=c.ElementList)
    involved_functions = c.ProxyAccessor(
        LogicalFunction,
        interaction.XT_CAP2ACT,
        aslist=c.ElementList,
        follow="involved",
    )
    involved_chains = c.ProxyAccessor(
        fa.FunctionalChain,
        interaction.XT_CAP2PROC,
        aslist=c.ElementList,
        follow="involved",
    )
    involved_components = c.ProxyAccessor(
        LogicalComponent,
        xtypes=ctx.CapabilityInvolvement,
        follow="involved",
        aslist=c.MixedElementList,
    )
    realized_capabilities = c.ProxyAccessor(
        ctx.Capability,
        interaction.XT_CAP_REAL,
        follow="targetElement",
        aslist=c.ElementList,
    )

    postcondition = c.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = c.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = c.ProxyAccessor(interaction.Scenario, aslist=c.ElementList)
    states = c.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CapabilityRealizationPkg(c.GenericElement):
    """A capability package that can hold capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = c.ProxyAccessor(CapabilityRealization, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(None)
class LogicalArchitecture(crosslayer.BaseArchitectureLayer):
    """Provides access to the LogicalArchitecture layer of the model."""

    root_component = c.AttributeMatcherAccessor(
        LogicalComponent,
        attributes={"is_actor": False},
        rootelem=LogicalComponentPkg,
    )
    root_function = c.ProxyAccessor(
        LogicalFunction, rootelem=LogicalFunctionPkg
    )

    function_package = c.ProxyAccessor(LogicalFunctionPkg)
    component_package = c.ProxyAccessor(LogicalComponentPkg)
    capability_package = c.ProxyAccessor(CapabilityRealizationPkg)

    all_functions = c.ProxyAccessor(
        LogicalFunction,
        aslist=c.ElementList,
        rootelem=LogicalFunctionPkg,
        deep=True,
    )
    all_capabilities = c.ProxyAccessor(
        CapabilityRealization, deep=True, aslist=c.ElementList
    )
    all_components = c.ProxyAccessor(  # maybe this should exclude .is_actor
        LogicalComponent, aslist=c.ElementList, deep=True
    )
    all_actors = property(
        lambda self: self._model.search(LogicalComponent).by_is_actor(True)
    )

    actor_exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=LogicalComponentPkg,
    )
    component_exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=[LogicalComponentPkg, LogicalComponent],
        deep=True,
    )

    all_function_exchanges = c.ProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[LogicalFunctionPkg, LogicalFunction],
        deep=True,
    )
    all_component_exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        deep=True,
    )

    diagrams = diagram.DiagramAccessor(
        "Logical Architecture", cacheattr="_MelodyModel__diagram_cache"
    )  # type: ignore[assignment]


c.set_accessor(
    ctx.Capability,
    "realizing_capabilities",
    c.CustomAccessor(
        CapabilityRealization,
        operator.attrgetter("_model.la.all_capabilities"),
        matchtransform=operator.attrgetter("realized_capabilities"),
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    ctx.SystemComponent,
    "realizing_logical_components",
    c.CustomAccessor(
        LogicalComponent,
        operator.attrgetter("_model.la.all_components"),
        matchtransform=operator.attrgetter("realized_components"),
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    ctx.SystemFunction,
    "realizing_logical_functions",
    c.CustomAccessor(
        LogicalFunction,
        operator.attrgetter("_model.la.all_functions"),
        matchtransform=operator.attrgetter("realized_system_functions"),
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    LogicalFunction,
    "packages",
    c.ProxyAccessor(
        LogicalFunctionPkg,
        aslist=c.ElementList,
    ),
)
c.set_self_references(
    (LogicalComponent, "components"),
    (LogicalComponentPkg, "packages"),
    (LogicalFunction, "functions"),
    (LogicalFunctionPkg, "packages"),
)



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Implementation of objects and relations for Functional Analysis

Composite Structure objects inheritance tree (taxonomy):

.. diagram:: [CDB] CompositeStructure [Taxonomy]

Composite Structure object-relations map (ontology):

.. diagram:: [CDB] CompositeStructure [Ontology]
"""

import operator

from capellambse.loader import xmltools

from .. import common as c
from . import capellacommon, fa, information

XT_DEPLOY_LINK = (
    "org.polarsys.capella.core.data.pa.deployment:PartDeploymentLink"
)
XT_PHYS_PATH_INV = "org.polarsys.capella.core.data.cs:PhysicalPathInvolvement"


@c.xtype_handler(None)
class Part(c.GenericElement):
    """A representation of a physical component"""

    _xmltag = "ownedParts"

    type = c.AttrProxyAccessor(c.GenericElement, "abstractType")

    deployed_parts: c.Accessor


@c.xtype_handler(None)
class Interface(c.GenericElement):
    """An interface."""


@c.xtype_handler(None)
class InterfacePkg(c.GenericElement):
    """A package that can hold interfaces and exchange items."""

    exchange_items = c.ProxyAccessor(
        information.ExchangeItem,
        aslist=c.ElementList,
    )
    interfaces = c.ProxyAccessor(Interface, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(None)
class PhysicalPort(c.GenericElement):
    """A physical port."""

    _xmltag = "ownedFeatures"

    owner = c.ParentAccessor(c.GenericElement)


@c.xtype_handler(None)
class PhysicalLink(PhysicalPort):
    """A physical link."""

    linkEnds = c.AttrProxyAccessor(
        PhysicalPort, "linkEnds", aslist=c.ElementList
    )
    exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        xtypes=fa.XT_COMP_EX_ALLOC,
        aslist=c.ElementList,
        follow="targetElement",
    )

    physical_paths: c.Accessor


@c.xtype_handler(None)
class PhysicalPath(c.GenericElement):
    """A physical path."""

    _xmltag = "ownedPhysicalPath"

    involved_items = c.ProxyAccessor(
        c.GenericElement,
        xtypes=XT_PHYS_PATH_INV,
        aslist=c.MixedElementList,
        follow="involved",
    )
    exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        xtypes=fa.XT_COMP_EX_ALLOC,
        aslist=c.ElementList,
        follow="targetElement",
    )

    @property
    def involved_links(self) -> c.ElementList[PhysicalLink]:
        return self.involved_items.by_type("PhysicalLink")


class Component(c.GenericElement):
    """A template class for components."""

    is_abstract = xmltools.BooleanAttributeProperty(
        "_element",
        "abstract",
        __doc__="Boolean flag for an abstract Component",
    )
    is_human = xmltools.BooleanAttributeProperty(
        "_element", "human", __doc__="Boolean flag for a human Component"
    )
    is_actor = xmltools.BooleanAttributeProperty(
        "_element", "actor", __doc__="Boolean flag for an actor Component"
    )

    owner = c.ParentAccessor(c.GenericElement)
    state_machines = c.ProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )
    ports = c.ProxyAccessor(fa.ComponentPort, aslist=c.ElementList)
    physical_ports = c.ProxyAccessor(PhysicalPort, aslist=c.ElementList)
    parts = c.ReferenceSearchingAccessor(Part, "type", aslist=c.ElementList)
    physical_paths = c.ProxyAccessor(PhysicalPath, aslist=c.ElementList)
    physical_links = c.ProxyAccessor(PhysicalLink, aslist=c.ElementList)
    exchanges = c.ReferenceSearchingAccessor(
        fa.ComponentExchange,
        "source.owner",
        "target.owner",
        aslist=c.ElementList,
    )


@c.xtype_handler(None)
class ComponentRealization(c.GenericElement):
    """A realization that links to a component."""

    _xmltag = "ownedComponentRealizations"


c.set_accessor(
    InterfacePkg,
    "packages",
    c.ProxyAccessor(InterfacePkg, aslist=c.ElementList),
)
c.set_accessor(
    Part,
    "deployed_parts",
    c.ProxyAccessor(
        Part,
        XT_DEPLOY_LINK,
        aslist=c.ElementList,
        follow="deployedElement",
        follow_abstract=False,
    ),
)
c.set_accessor(
    PhysicalLink,
    "physical_paths",
    c.CustomAccessor(
        PhysicalPath,
        operator.attrgetter("_model.pa.all_physical_paths"),
        matchtransform=operator.attrgetter("involved_items"),
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    fa.ComponentExchange,
    "allocating_physical_link",
    c.ReferenceSearchingAccessor(PhysicalLink, "exchanges"),
)
c.set_accessor(
    fa.ComponentExchange,
    "allocating_physical_path",
    c.ReferenceSearchingAccessor(PhysicalPath, "exchanges"),
)



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Parser entry point for semantic elements.

Semantic elements have a ``<target>`` which references the represented
object in the ``.melodymodeller`` or ``.melodyfragment`` file.  The
melody file is the source of truth for all attributes that are not
specific to a single diagram, i.e. basically everything except position,
size and styling.
"""
from __future__ import annotations

import collections.abc as cabc
import logging
import typing as t

from lxml import etree

import capellambse  # pylint: disable=unused-import  # used in typing
from capellambse import aird

from . import _box_factories
from . import _common as c
from . import _edge_factories

LOGGER = logging.getLogger(__name__)
NO_RENDER_XMT = frozenset(
    {
        "diagram:DNodeListElement",
    }
)


def from_xml(ebd: c.ElementBuilder) -> aird.DiagramElement:
    """Deserialize a semantic element."""
    uid = ebd.data_element.attrib.get("element")
    if uid is None:
        raise c.SkipObject()

    diag_element = ebd.melodyloader.follow_link(ebd.data_element, uid)
    if diag_element.get(c.ATT_XMT) in NO_RENDER_XMT:
        raise c.SkipObject()

    target = next(diag_element.iterchildren("target"), None)
    if target is None:
        raise c.SkipObject()
    needed_target_attrib = frozenset({"href", c.ATT_XMT})
    actual_target_attrib = set(target.attrib) & needed_target_attrib
    if actual_target_attrib != needed_target_attrib:
        LOGGER.error(
            "Missing required attributes %s from target of %r",
            ", ".join(actual_target_attrib ^ needed_target_attrib),
            uid,
        )
        raise c.SkipObject()

    styleclass = target.attrib[c.ATT_XMT]
    try:
        styleclass = styleclass.split(":")[1]
    except IndexError:
        raise ValueError(f"Invalid target type {styleclass}") from None

    try:
        styleclass, drawtype = STYLECLASS_LOOKUP[styleclass]
    except KeyError:
        drawtype = _GENERIC_FACTORIES

    sem_elms = list(diag_element.iterchildren("semanticElements"))
    melodyobjs: list[etree._Element] = []
    for sem_elm in sem_elms:
        sem_href = sem_elm.attrib["href"]
        try:
            sem_obj = ebd.melodyloader.follow_link(sem_elm, sem_href)
        except KeyError:
            LOGGER.warning(
                "Referenced semantic element %r does not exist", sem_href
            )
        melodyobjs.append(sem_obj)
    assert melodyobjs, "No valid semantic element references could be found"
    if melodyobjs[0] is None:
        raise c.SkipObject()

    seb = c.SemanticElementBuilder(
        target_diagram=ebd.target_diagram,
        diagram_tree=ebd.diagram_tree,
        data_element=ebd.data_element,
        melodyloader=ebd.melodyloader,
        fragment=ebd.fragment,
        styleclass=styleclass,
        diag_element=diag_element,
        melodyobjs=melodyobjs,
    )
    return drawtype(seb)


class FactorySelector:
    """Selects a factory based on whether it's for a box or an edge."""

    __slots__ = ("box", "edge")

    def __init__(
        self,
        box: cabc.Callable[[c.SemanticElementBuilder], aird.Box],
        edge: cabc.Callable[[c.SemanticElementBuilder], aird.Edge],
    ) -> None:
        self.box = box
        self.edge = edge

    def __call__(self, seb: c.SemanticElementBuilder) -> aird.DiagramElement:
        factory: cabc.Callable[[c.SemanticElementBuilder], aird.DiagramElement]
        if seb.data_element.tag == "children":
            factory = self.box
        elif seb.data_element.tag == "edges":
            factory = self.edge
        else:
            raise ValueError(f"Unknown element type {seb.data_element.tag}")
        return factory(seb)


_GENERIC_FACTORIES = FactorySelector(
    _box_factories.generic_factory, _edge_factories.generic_factory
)
#: This dictionary implements a mapping between the Capella internal
#: types and the types used in the drawing interface protocol.
#: The dictionary keys are the Capella types, the values are a tuple of
#: the style class and the specific factory function.
#:
#: If a key is not found in this dictionary, its value is assumed to be
#: a tuple of its key and the ``_GENERIC_FACTORIES``.
SemanticDeserializer = t.Callable[
    [c.SemanticElementBuilder], "capellambse.aird.DiagramElement"
]
STYLECLASS_LOOKUP: dict[str, tuple[str | None, SemanticDeserializer]]
STYLECLASS_LOOKUP = {
    "AbstractCapabilityInclude": (
        "AbstractCapabilityInclude",
        _edge_factories.include_extend_factory,
    ),
    "AbstractCapabilityExtend": (
        "AbstractCapabilityExtend",
        _edge_factories.include_extend_factory,
    ),
    "Association": (
        "Association",
        _edge_factories.association_factory,
    ),
    "CapellaIncomingRelation": (
        "RequirementRelation",
        _edge_factories.req_relation_factory,
    ),
    "CapellaOutgoingRelation": (
        "RequirementRelation",
        _edge_factories.req_relation_factory,
    ),
    "ChoicePseudoState": (
        "ChoicePseudoState",
        _box_factories.pseudo_symbol_factory,
    ),
    "Class": ("Class", _box_factories.class_factory),
    "ComponentPort": ("CP", _box_factories.component_port_factory),
    "Constraint": (
        "Constraint",
        FactorySelector(
            _box_factories.constraint_factory,
            _edge_factories.constraint_factory,
        ),
    ),
    "ControlNode": ("ControlNode", _box_factories.control_node_factory),
    "ExchangeItemElement": (
        "ExchangeItemElement",
        _edge_factories.eie_factory,
    ),
    "Enumeration": ("Enumeration", _box_factories.enumeration_factory),
    "EnumerationLiteral": (
        None,
        c.SkipObject.raise_,
    ),  # handled as part of `Enumeration`
    "Entity": (
        "Entity",
        FactorySelector(
            _box_factories.generic_factory, _edge_factories.labelless_factory
        ),
    ),
    "FunctionInputPort": ("FIP", _GENERIC_FACTORIES),
    "FunctionOutputPort": ("FOP", _GENERIC_FACTORIES),
    "ForkPseudoState": (
        "ForkPseudoState",
        _box_factories.pseudo_symbol_factory,
    ),
    "Mode": ("Mode", _box_factories.statemode_factory),
    "State": ("State", _box_factories.statemode_factory),
    "StateTransition": (
        "StateTransition",
        _edge_factories.state_transition_factory,
    ),
    "SystemComponent": ("SystemComponent", _box_factories.part_factory),
    "OperationalActor": (
        "OperationalActor",
        FactorySelector(
            _box_factories.generic_factory, _edge_factories.labelless_factory
        ),
    ),
    "Part": (None, _box_factories.part_factory),
    "Region": ("Region", _box_factories.region_factory),
    "PortAllocation": (
        "PortAllocation",
        _edge_factories.port_allocation_factory,
    ),
    "Property": (None, c.SkipObject.raise_),  # Handled as part of `Class`
    "Service": (None, c.SkipObject.raise_),  # Handled as part of `Class`
    "SequenceLink": ("SequenceLink", _edge_factories.sequence_link_factory),
    "PhysicalPort": ("PP", _GENERIC_FACTORIES),
    "Requirement": ("Requirement", _box_factories.requirements_box_factory),
    # FunctionalChains
    "FunctionalChainInvolvementFunction": (
        "FunctionalChainInvolvementFunction",
        _box_factories.fcif_factory,
    ),
    "FunctionalChainInvolvementLink": (
        "FunctionalChainInvolvementLink",
        _edge_factories.fcil_factory,
    ),
    "FunctionalChainReference": (
        "FunctionalChainReference",
        _box_factories.generic_stacked_factory,
    ),
}



# Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Helps loading Capella models (including fragmented variants)."""
from __future__ import annotations

__all__ = [
    "FragmentType",
    "MelodyLoader",
]

import collections
import collections.abc as cabc
import contextlib
import enum
import itertools
import logging
import operator
import os.path
import pathlib
import re
import sys
import typing as t
import urllib.parse
import uuid

from lxml import etree

import capellambse
from capellambse import helpers
from capellambse.loader import exs, filehandler
from capellambse.loader.filehandler import localfilehandler
from capellambse.loader.modelinfo import ModelInfo

LOGGER = logging.getLogger(__name__)
VISUAL_EXTS = frozenset(
    {
        ".aird",
        ".airdfragment",
    }
)
SEMANTIC_EXTS = frozenset(
    {
        ".capella",
        ".capellafragment",
        ".melodyfragment",
        ".melodymodeller",
    }
)
VALID_EXTS = VISUAL_EXTS | SEMANTIC_EXTS | {".afm"}
ERR_BAD_EXT = "Model file {} has an unsupported extension: {}"

IDTYPES = frozenset({"id", "uid", "xmi:id"})
IDTYPES_RESOLVED = frozenset(helpers.resolve_namespace(t) for t in IDTYPES)
RE_VALID_ID = re.compile(
    r"([a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12})"
)
CAP_VERSION = re.compile(r"Capella_Version_([\d.]+)")
CROSS_FRAGMENT_LINK = re.compile(
    r"""
    ^
    (?:
        (?:
            (?:(?P<xtype>[^ #]+)\ )?
            (?P<fragment>[^ #]+)
        )?
        \#
    )?
    (?P<uuid>[^ #]+)
    $
    """,
    re.VERBOSE,
)


def _verify_extension(filename: pathlib.PurePosixPath) -> None:
    """Check whether ``filename`` has a valid extension."""
    file = pathlib.PurePosixPath(filename)
    if file.suffix not in VALID_EXTS:
        raise TypeError(ERR_BAD_EXT.format(file, file.suffix))


def _find_refs(root: etree._Element) -> cabc.Iterable[str]:
    return itertools.chain(
        (x.split("#")[0] for x in root.xpath(".//referencedAnalysis/@href")),
        root.xpath(".//semanticResources/text()"),
    )


def _unquote_ref(ref: str) -> str:
    ref = urllib.parse.unquote(ref)
    prefix = "platform:/resource/"
    if ref.startswith(prefix):
        ref = ref.replace(prefix, "../")
    return ref


class FragmentType(enum.Enum):
    """The type of an XML fragment."""

    SEMANTIC = enum.auto()
    VISUAL = enum.auto()
    OTHER = enum.auto()


class MissingResourceLocationError(KeyError):
    """Raised when a model needs an additional resource location."""


class ResourceLocationManager(dict):
    def __missing__(self, key: str) -> t.NoReturn:
        raise MissingResourceLocationError(key)


class ModelFile:
    """Represents a single file in the model (i.e. a fragment)."""

    __xtypecache: dict[str, list[etree._Element]]
    __idcache: dict[str, etree._Element]
    __hrefsources: dict[str, etree._Element]

    @property
    def fragment_type(self) -> FragmentType:
        if self.filename.suffix in SEMANTIC_EXTS:
            return FragmentType.SEMANTIC
        elif self.filename.suffix in VISUAL_EXTS:
            return FragmentType.VISUAL
        else:
            return FragmentType.OTHER

    def __init__(
        self, filename: pathlib.PurePosixPath, handler: filehandler.FileHandler
    ) -> None:
        self.filename = filename
        self.filehandler = handler
        _verify_extension(filename)

        with handler.open(filename) as f:
            self.tree = etree.parse(
                f, etree.XMLParser(remove_blank_text=True, huge_tree=True)
            )

        self.root = self.tree.getroot()
        self.idcache_rebuild()

    def __getitem__(self, key: str) -> etree._Element:
        return self.__idcache[key]

    def idcache_index(self, subtree: etree._Element) -> None:
        """Index the IDs of ``subtree``."""
        for elm in subtree.iter():
            xtype = helpers.xtype_of(elm)
            if xtype is not None:
                self.__xtypecache[xtype].append(elm)

            for idtype in IDTYPES_RESOLVED:
                elm_id = elm.get(idtype, None)
                if elm_id is None:
                    continue
                self.__idcache[elm_id] = elm

            href = elm.get("href")
            if href is not None:
                self.__hrefsources[href.split("#")[-1]] = elm

    def idcache_remove(self, source: str | etree._Element) -> None:
        """Remove the ID or all IDs below the source from the ID cache."""
        if isinstance(source, str):
            try:
                del self.__idcache[source]
            except KeyError:
                pass

        else:
            for elm in source.iter():
                xtype = helpers.xtype_of(elm)
                if xtype:
                    self.__xtypecache[xtype].remove(elm)
                for idtype in IDTYPES_RESOLVED:
                    elm_id = elm.get(idtype, None)
                    if elm_id is None:
                        continue

                    try:
                        del self.__idcache[elm_id]
                    except KeyError:
                        pass
                href = elm.get("href")
                if href is not None:
                    del self.__hrefsources[href.split("#")[-1]]

    def idcache_rebuild(self) -> None:
        """Invalidate and rebuild this file's ID cache."""
        LOGGER.debug("Indexing file %s...", self.filename)
        self.__xtypecache = collections.defaultdict(list)
        self.__idcache = {}
        self.__hrefsources = {}
        self.idcache_index(self.root)
        LOGGER.debug("Cached %d element IDs", len(self.__idcache))

    def idcache_reserve(self, new_id: str) -> None:
        """Reserve the given ID for an element to be inserted later."""
        self.__idcache[new_id] = None

    def iterall_xt(
        self, xtypes: cabc.Container[str]
    ) -> cabc.Iterator[etree._Element]:
        """Iterate over all elements in this tree by ``xsi:type``."""
        for xtype, elms in self.__xtypecache.items():
            if xtype in xtypes:
                yield from elms

    def write_xml(
        self,
        filename: pathlib.PurePosixPath,
        encoding: str = "utf-8",
    ) -> None:
        """Write this file's XML into the file specified by ``path``."""
        LOGGER.debug("Saving tree %r to file %s", self, filename)
        if filename.suffix in {
            ".capella",
            ".capellafragment",
            ".melodyfragment",
            ".melodymodeller",
        }:
            line_length = exs.LINE_LENGTH
        else:
            line_length = sys.maxsize
        with self.filehandler.open(filename, "wb") as file:
            exs.write(
                self.tree, file, encoding=encoding, line_length=line_length
            )

    def unfollow_href(self, element_id: str) -> etree._Element:
        """ "Unfollow" a fragment link and return the placeholder element.

        If the given UUID is not linked to from this file, None is
        returned.
        """
        return self.__hrefsources.get(element_id)


class MelodyLoader:
    """Facilitates extensive access to Polarsys / Capella projects."""

    def __init__(
        self,
        path: str | os.PathLike | filehandler.FileHandler,
        entrypoint: str | pathlib.PurePosixPath | None = None,
        *,
        resources: cabc.Mapping[
            str,
            filehandler.FileHandler | str | os.PathLike | dict[str, t.Any],
        ]
        | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Construct a MelodyLoader.

        Parameters
        ----------
        path
            The ``path`` argument to the primary file handler, or the
            primary file handler itself.
        entrypoint
            The entry point into the model, i.e. the top-level ``.aird``
            file. This must be located within the primary file handler.
        resources
            Additional file handler instances that provide library
            resources that are referenced from the model.
        kwargs
            Additional arguments to the primary file handler, if
            necessary.
        """
        if isinstance(path, filehandler.FileHandler):
            handler = path
        else:
            handler = filehandler.get_filehandler(path, **kwargs)
        self.resources = ResourceLocationManager({"\0": handler})
        for resname, reshdl in (resources or {}).items():
            if not resname:
                raise ValueError("Empty resource name")
            if "/" in resname or "\0" in resname:
                raise ValueError(f"Invalid resource name: {resname!r}")

            if isinstance(reshdl, (str, os.PathLike)):
                self.resources[resname] = filehandler.get_filehandler(reshdl)
            elif isinstance(reshdl, cabc.Mapping):
                self.resources[resname] = filehandler.get_filehandler(**reshdl)
            else:
                self.resources[resname] = reshdl
        self.entrypoint = self.__derive_entrypoint(entrypoint)

        self.trees: dict[pathlib.PurePosixPath, ModelFile] = {}
        self.__load_referenced_files(
            pathlib.PurePosixPath("\0", self.entrypoint)
        )

    @property
    def filehandler(self) -> filehandler.FileHandler:
        return self.resources["\0"]

    def __derive_entrypoint(
        self, entrypoint: str | pathlib.PurePosixPath | None
    ) -> pathlib.PurePosixPath:
        if entrypoint:
            return helpers.normalize_pure_path(entrypoint)

        if isinstance(self.filehandler, localfilehandler.LocalFileHandler):
            basedir = self.filehandler.path
            assert isinstance(basedir, pathlib.Path)
            self.filehandler.path = basedir.parent
            return helpers.normalize_pure_path(basedir.name)

        raise ValueError("This type of file handler needs an ``entrypoint``")

    def __load_referenced_files(
        self, resource_path: pathlib.PurePosixPath
    ) -> None:
        if resource_path in self.trees:
            return

        handler = self.resources[resource_path.parts[0]]
        filename = pathlib.PurePosixPath(*resource_path.parts[1:])
        frag = ModelFile(filename, handler)
        self.trees[resource_path] = frag
        for ref in _find_refs(frag.root):
            ref_name = helpers.normalize_pure_path(
                _unquote_ref(ref), base=resource_path.parent
            )
            self.__load_referenced_files(ref_name)

    def save(self, **kw: t.Any) -> None:
        """Save all model files back to their original locations.

        Parameters
        ----------
        kw
            Additional keyword arguments accepted by the file handler in
            use. Please see the respective documentation for more info.

        See Also
        --------
        capellambse.loader.filehandler.localfilehandler.LocalFileHandler.write_transaction :
            Accepted ``**kw`` when using local directories
        capellambse.loader.filehandler.gitfilehandler.GitFileHandler.write_transaction :
            Accepted ``**kw`` when using ``git://`` and similar URLs
        """
        LOGGER.debug("Saving model %r", self.get_model_info().title)
        with self.filehandler.write_transaction(**kw) as unsupported_kws:
            if unsupported_kws:
                LOGGER.warning(
                    "Ignoring unsupported transaction parameters: %s",
                    ", ".join(repr(k) for k in unsupported_kws),
                )
            for fname, tree in self.trees.items():
                resname = fname.parts[0]
                fname = pathlib.PurePosixPath(*fname.parts[1:])
                if resname != "\0":
                    continue

                tree.write_xml(fname)

    def idcache_index(self, subtree: etree._Element) -> None:
        """Index the IDs of ``subtree``.

        This method must be called after adding ``subtree`` to the XML
        tree.
        """
        try:
            _, tree = self._find_fragment(subtree)
        except ValueError:
            raise ValueError(
                "Call idcache_index() after adding the subtree"
            ) from None

        tree.idcache_index(subtree)

    def idcache_remove(self, subtree: etree._Element) -> None:
        """Remove the ``subtree`` from the ID cache.

        This method must be called before actually removing ``subtree``
        from the XML tree.
        """
        try:
            _, tree = self._find_fragment(subtree)
        except ValueError:
            raise ValueError(
                "Call idcache_remove() before removing the subtree"
            ) from None

        tree.idcache_remove(subtree)

    def idcache_rebuild(self) -> None:
        r"""Rebuild the ID caches of all :class:`ModelFile`\ s."""
        for tree in self.trees.values():
            tree.idcache_rebuild()

    def generate_uuid(
        self, parent: etree._Element, *, want: str | None = None
    ) -> str:
        """Generate a unique UUID for a new child of ``parent``.

        The generated ID is guaranteed to be unique across all currently
        loaded fragments.

        Parameters
        ----------
        want
            Try this UUID first, and use it if it satisfies all other
            constraints. If it does not satisfy all constraints (e.g. it
            would be non-unique), a random UUID will be generated as
            normal.
        """

        def idstream() -> t.Iterator[str]:
            if want and RE_VALID_ID.fullmatch(want):
                yield want
            while True:
                yield str(uuid.uuid4())

        _, tree = self._find_fragment(parent)

        for new_id in idstream():
            try:
                self[new_id]
            except KeyError:
                tree.idcache_reserve(new_id)
                return new_id
        assert False

    @contextlib.contextmanager
    def new_uuid(
        self, parent: etree._Element, *, want: str | None = None
    ) -> cabc.Generator[str, None, None]:
        """Context Manager around :meth:`generate_uuid()`.

        This context manager yields a newly generated model-wide unique
        UUID that can be inserted into a new element during the ``with``
        block.  It tries to keep the ID cache consistent in some harder
        to manage edge cases, like exceptions being thrown.
        Additionally it checks that the generated UUID was actually used
        in the tree; not using it before the ``with`` block ends is an
        error and provokes an Exception.

        .. note:: You still need to call :meth:`idcache_index()` on the
            newly inserted element!

        Example usage::

            >>> with ldr.new_uuid(parent_elm) as obj_id:
            ...     child_elm = parent_elm.makeelement("ownedObjects")
            ...     child_elm.set("id", obj_id)
            ...     parent_elm.append(child_elm)
            ...     ldr.idcache_index(child_elm)

        If you intend to reserve a UUID that should be inserted later,
        use :meth:`generate_uuid()` directly.

        Parameters
        ----------
        parent
            The parent element below which the new UUID will be used.
        """
        _, tree = self._find_fragment(parent)
        new_uuid = self.generate_uuid(parent, want=want)

        def cleanup_after_failure() -> None:
            tree.idcache_remove(new_uuid)
            for child in parent:
                for id_attr in IDTYPES_RESOLVED:
                    if child.get(id_attr) == new_uuid:
                        parent.remove(child)

        try:
            yield new_uuid
        except BaseException:
            cleanup_after_failure()
            raise

        if self[new_uuid] is None:
            cleanup_after_failure()
            raise RuntimeError("New UUID was requested but never used")

    def xpath(
        self,
        query: str | etree.XPath,
        *,
        namespaces: cabc.Mapping[str, str] | None = None,
        roots: etree._Element | cabc.Iterable[etree._Element] | None = None,
    ) -> list[etree._Element]:
        """Run an XPath query on all fragments.

        Parameters
        ----------
        query
            The XPath query
        namespaces
            Namespaces used in the query.  Defaults to all known
            namespaces.
        roots
            A list of XML elements to use as roots for the query.
            Defaults to all tree roots.
        """
        return list(
            map(
                operator.itemgetter(1),
                self.xpath2(query, namespaces=namespaces, roots=roots),
            )
        )

    def xpath2(
        self,
        query: str | etree.XPath,
        *,
        namespaces: cabc.Mapping[str, str] | None = None,
        roots: etree._Element | cabc.Iterable[etree._Element] | None = None,
    ) -> list[tuple[pathlib.PurePosixPath, etree._Element]]:
        """Run an XPath query and return the fragments and elements.

        The tuples have the fragment where the match was found as first
        element, and the LXML element as second one.

        Parameters
        ----------
        query
            The XPath query
        namespaces
            Namespaces used in the query.  Defaults to all known
            namespaces.
        roots
            A list of XML elements to use as roots for the query.
            Defaults to all tree roots.
        """
        if namespaces is None:
            namespaces = capellambse.NAMESPACES

        def follow_href(
            tree: pathlib.PurePosixPath, match: etree._Element
        ) -> tuple[pathlib.PurePosixPath, etree._Element]:
            if href := match.get("href"):
                match = self.follow_link(match, href)
                return (self.find_fragment(match), match)
            return (tree, match)

        if not isinstance(query, etree.XPath):
            query = etree.XPath(query, namespaces=namespaces)

        if roots is None:
            roots = [(k, t.root) for k, t in self.trees.items()]
        elif isinstance(roots, etree._Element):
            roots = [(self._find_fragment(roots)[0], roots)]
        elif isinstance(roots, cabc.Iterable):
            roots = [(self._find_fragment(r)[0], r) for r in roots]
        else:
            raise TypeError(
                f"`roots` must be an XML element or a list thereof, not {type(roots).__name__}"
            )

        ret = []
        for fragment, tree in roots:
            ret += [follow_href(fragment, elem) for elem in query(tree)]
        return ret

    def find_by_xsi_type(
        self,
        *xsi_types: str,
        roots: etree._Element | cabc.Iterable[etree._Element] = None,
    ) -> list[etree._Element]:
        r"""Find all elements matching any of the given ``xsi:type``\ s.

        Parameters
        ----------
        xsi_types
            ``xsi:type`` strings to match, for example
            "org.polarsys.capella.core.data.cs:InterfacePkg"
        roots
            A list of XML elements to use as roots for the query.
            Defaults to all tree roots.
        """
        if roots is None:
            return list(self.iterall_xt(*xsi_types))
        elif isinstance(roots, etree._Element):
            roots = (roots,)

        xtset = self._nonempty_hashset(xsi_types)
        del xsi_types

        return [
            i
            for i in itertools.chain.from_iterable(roots)
            if (xt := helpers.xtype_of(i)) is not None and xt in xtset
        ]

    def iterall(self, *tags: str) -> cabc.Iterator[etree._Element]:
        """Iterate over all elements in all trees by tags.

        Parameters
        ----------
        tags
            Optionally restrict the iterator to the given tags.
        """
        return itertools.chain.from_iterable(
            t.root.iter(*tags) for t in self.trees.values()
        )

    def iterall_xt(
        self,
        *xtypes: str,
        trees: cabc.Container[pathlib.PurePosixPath] | None = None,
    ) -> cabc.Iterator[etree._Element]:
        r"""Iterate over all elements in all trees by ``xsi:type``\ s.

        Parameters
        ----------
        xtypes
            Optionally restrict the iterator to these ``xsi:type``\ s
        trees
            Optionally restrict the iterator to elements that reside in
            any of the named trees.
        """
        xtset = self._nonempty_hashset(xtypes)
        if trees is None:
            files: cabc.Iterable[ModelFile] = self.trees.values()
        else:
            files = (v for k, v in self.trees.items() if k in trees)
        return itertools.chain.from_iterable(
            map(operator.methodcaller("iterall_xt", xtset), files)
        )

    def iterdescendants(
        self,
        root_elm: etree._Element,
        *tags: str,
    ) -> cabc.Iterator[etree._Element]:
        """Iterate over all descendants of ``root_elm``.

        This method will follow links into different fragment files and
        yield those elements as if they were part of the origin subtree.

        Parameters
        ----------
        root_elm
            The root element of the tree
        tags
            Element tags to filter for
        """
        tagset = self._nonempty_hashset(tags)
        it_stack = [root_elm.iterdescendants()]
        while it_stack:
            realelm = None
            try:
                elm = next(it_stack[-1])
            except StopIteration:
                it_stack.pop()
                continue

            if "href" in elm.attrib:  # Follow into the fragment
                href = elm.attrib["href"].split()[-1]
                realelm = self[href]
                it_stack.append(realelm.iterdescendants())

            if elm.tag in tagset:
                if realelm is not None:
                    yield realelm
                else:
                    yield elm

    def iterdescendants_xt(
        self,
        element: etree._Element,
        *xtypes: str,
    ) -> cabc.Iterator[etree._Element]:
        r"""Iterate over all descendants of ``element`` by ``xsi:type``.

        This method will follow links into different fragment files and
        yield those elements as if they were part of the origin subtree.

        Parameters
        ----------
        element
            The root element of the tree
        xtypes
            Elements that have any of these ``xsi:type``\ s will be
            yielded.  If nothing is given here, all elements will be
            yielded.
        """
        xtset = self._nonempty_hashset(xtypes)
        return (
            i
            for i in self.iterdescendants(element)
            if helpers.xtype_of(i) in xtset
        )

    def iterancestors(
        self,
        element: etree._Element,
        *tags: str,
    ) -> cabc.Iterator[etree._Element]:
        """Iterate over the ancestors of ``element``.

        This method will follow fragment links back to the origin point.
        """
        tagset = self._nonempty_hashset(tags)
        visited_elements = []  # Basic protection against reference loops
        while True:
            parent = element.getparent()
            if parent is None:
                possible_sources = []
                for idtype in IDTYPES_RESOLVED:
                    try:
                        possible_sources.append(
                            self._unfollow_href(element.attrib[idtype])
                        )
                    except KeyError:
                        pass
                assert 0 <= len(possible_sources) <= 1
                if possible_sources:
                    parent = possible_sources[0].getparent()
                else:
                    break
            element = parent

            assert element not in visited_elements
            visited_elements.append(element)
            if element.tag in tagset:
                yield element

    def iterchildren_xt(
        self, element: etree._Element, *xtypes: str
    ) -> cabc.Iterator[etree._Element]:
        r"""Iterate over the children of ``element``.

        If ``xsi:type``\ s are given in ``xtypes``, restrict yielded
        elements to the ones with matching type.
        """
        xtset = self._nonempty_hashset(xtypes)
        for child in element.iterchildren():
            child = self._follow_href(child)
            if helpers.xtype_of(child) in xtset:
                yield child

    def create_link(
        self,
        from_element: etree._Element,
        to_element: etree._Element,
    ) -> str:
        """Create a link to ``to_element`` from ``from_element``.

        The resulting link is valid for insertion into ``from_element``
        or any of its children, and will reference ``to_element`` either
        in the same or a different fragment.  It has one of the
        following three formats:

        1.  Within the same fragment: ``#UUID``
        2.  Across different fragments: ``target_xsitype relpath#UUID``
        3.  Across different fragments, if the target does not have an
            ``xsi:type``: ``relpath#UUID``

        Parameters
        ----------
        from_element
            The source element of the link.
        to_element
            The target element of the link.
        """
        to_uuids = set(to_element.keys()) & IDTYPES_RESOLVED
        try:
            to_uuid = next(iter(to_uuids))
        except StopIteration:
            raise ValueError(
                "to_element does not have a known ID attribute"
            ) from None
        to_uuid = to_element.get(to_uuid)

        # Find the fragments corresponding to each tree
        from_fragment, _ = self._find_fragment(from_element)
        to_fragment, _ = self._find_fragment(to_element)
        assert from_fragment and to_fragment

        if from_fragment == to_fragment:
            return f"#{to_uuid}"

        to_fragment = pathlib.PurePosixPath(
            os.path.relpath(to_fragment, from_fragment.parent)
        )
        link = urllib.parse.quote(str(to_fragment))
        to_type = helpers.xtype_of(to_element)
        if to_type is not None:
            return f"{to_type} {link}#{to_uuid}"
        return f"{link}#{to_uuid}"

    def follow_link(
        self,
        from_element: etree._Element | None,
        link: str,
    ) -> etree._Element:
        """Follow a single link and return the target element.

        The link is considered relative to the ``from_element``, if
        given. The accepted formats are as follows::

            xtype fragment#UUID
            fragment#UUID
            #UUID
            UUID

        Where:

        *   ``xtype`` is the target element's ``xsi:type``.  If given
            and it does not match, an exception is raised.
        *   ``fragment`` is the fragment file in which to find the
            target element.  If not given, first the source fragment
            will be searched, and if unsuccessful, the search will be
            extended to all fragments.
        *   ``UUID`` is the unique ``id`` of the target element.

        Raises
        ------
        ValueError
            If the link is malformed
        FileNotFoundError
            If the target fragment is not loaded (only applicable if
            ``from_element`` is not None and ``fragment`` is part of the
            link)
        RuntimeError
            If the expected ``xsi:type`` does not match the actual
            ``xsi:type`` of the found target
        KeyError
            If the target cannot be found
        """
        linkmatch = CROSS_FRAGMENT_LINK.fullmatch(link)
        if not linkmatch:
            raise ValueError(f"Malformed link: {link!r}")
        xtype, fragment, ref = linkmatch.groups()
        if fragment is not None:
            fragment = urllib.parse.unquote(_unquote_ref(fragment))

        def find_trees(
            from_element: etree._Element | None,
            fragment: pathlib.PurePosixPath | None,
        ) -> cabc.Iterable[ModelFile]:
            if fragment and from_element is None:
                return (
                    v for k, v in self.trees.items() if k.name == fragment.name
                )
            elif fragment:
                sourcefragment = self._find_fragment(from_element)[0]
                fragment = capellambse.helpers.normalize_pure_path(
                    fragment, base=sourcefragment.parent
                )
                try:
                    return [self.trees[fragment]]
                except KeyError:  # pragma: no cover
                    raise FileNotFoundError(
                        f"Fragment not loaded: {fragment}"
                    ) from None
            else:
                sourcefragment = pathlib.PurePosixPath("/")
                if from_element is not None:
                    sourcefragment = self._find_fragment(from_element)[0]
                return map(
                    operator.itemgetter(1),
                    sorted(
                        self.trees.items(),
                        key=lambda tree: tree[0].name != sourcefragment.name,
                    ),
                )

        trees = find_trees(
            from_element,
            pathlib.PurePosixPath(fragment) if fragment else None,
        )

        matches = []
        for tree in trees:
            try:
                matches.append(tree[ref])
            except KeyError:
                pass
        if not matches:
            raise KeyError(link)
        if len(matches) > 1:
            raise KeyError(f"Ambiguous reference: {link!r}")
        if xtype is not None:
            actual_xtype = helpers.xtype_of(matches[0])
            if actual_xtype != xtype:
                raise TypeError(
                    f"Bad XML: Expected a {xtype!r}, got {actual_xtype!r}"
                )
        return matches[0]

    def follow_links(
        self,
        from_element: etree._Element | None,
        links: str,
    ) -> list[etree._Element | None]:
        """Follow multiple links and return all results as list.

        The format for an individual link is the same as accepted by
        :meth:`follow_link`.  Multiple links must be given as a single
        space-separated string.

        If any target cannot be found, ``None`` will be inserted at that
        point in the returned list.

        Raises
        ------
        ValueError
            If any link is malformed.
        RuntimeError
            If any expected ``xsi:type`` does not match the actual
            ``xsi:type`` of the found target.
        """
        targets = []
        next_xtype = None
        for part in links.split():
            if "#" not in part:
                if next_xtype is not None:
                    raise ValueError(f"Malformed link definition: {links}")
                next_xtype = part
                continue

            if next_xtype is not None:
                part = f"{next_xtype} {part}"
            try:
                targets.append(self.follow_link(from_element, part))
            except KeyError:
                targets.append(None)
        return targets

    def _find_fragment(
        self, element: etree._Element
    ) -> tuple[pathlib.PurePosixPath, ModelFile]:
        root = collections.deque(
            itertools.chain([element], element.iterancestors()), 1
        )[0]
        for fragment, tree in self.trees.items():
            if tree.root is root:
                return (fragment, tree)
        raise ValueError("Element is not contained in any fragment")

    def _follow_href(self, element: etree._Element) -> etree._Element:
        href = element.get("href")
        if href is None:
            return element
        return self[href]

    def _unfollow_href(self, element_id: str) -> etree._Element:
        for tree in self.trees.values():
            element = tree.unfollow_href(element_id)
            if element is not None:
                return element
        raise KeyError(element_id)

    def find_fragment(self, element: etree._Element) -> pathlib.PurePosixPath:
        """Find the name of the fragment that contains ``element``."""
        return self._find_fragment(element)[0]

    def __getitem__(self, key: str) -> etree._Element:
        """Search all loaded fragments for the given UUID."""
        return self.follow_link(None, key)

    @staticmethod
    def _nonempty_hashset(tags: tuple[str, ...]) -> cabc.Container[str]:
        if not tags:
            return helpers.EverythingContainer()
        try:
            return set(tags)
        except TypeError:
            return tags

    def get_model_info(self) -> ModelInfo:
        """Return the Capella version found in the leading comment."""
        info = self.filehandler.get_model_info()
        info.capella_version = "UNKNOWN"

        for fragment, tree in self.trees.items():
            if fragment.name.endswith((".capella", ".melodymodeller")):
                semantic_tree = tree
                break
        else:
            LOGGER.warning(
                "Cannot find capella version: No main semantic tree found!"
            )
            return info

        try:
            comment = semantic_tree.tree.xpath("/comment()")
            info.capella_version = CAP_VERSION.match(comment[0].text).group(1)  # type: ignore[union-attr]
        except (AttributeError, IndexError):
            LOGGER.warning(
                "Cannot find Capella version: No version comment found"
            )

        return info
